/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <dlfcn.h>
#include <event.h>
#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sysexits.h>

#include "tun_dev.h"
#include "mtund.h"

/* max transfered unit - encapsulated packet size */
#define MTU 1500
/*
 * probes ad keep-alive pings
 *
 * PING_INTERVAL    - how often to send pings
 * PING_FAIL_PLUGIN - how many can fail before the plugin is declared broken
 * PING_FAIL_PLUGIN - how many can fail before the client is declared
 *                    dead by the server and its ID reclaimed
 *
 * The counter for failed pings is reset whenever a ping reply is
 * received or the plugin is changed.
 */
#define PING_INTERVAL 5
#define PING_FAIL_PLUGIN 3
#define PING_FAIL_COMPLETE 10

#define min(a,b) ( (a>b) ? b : a )


/* DATA TYPES */
/* fragment header */
struct frag_hdr {
	uint8_t  dispatch;
	uint16_t id;	/* ID of the packet. same in each fragment
			 * belonging to the same packet */
	uint16_t size;	/* length of the whole packet, same in each fragment */
	uint16_t offset;/* fragment offset (in bytes) from the beginning
			 * of the packet */
};

/* info about a packet being reassembled from fragments */
struct frag_info {
	uint16_t id;		/* ID of the packet. same in each fragment
				 * belonging to the same packet */
	uint16_t size;		/* length of the whole packet, same
				 * in each fragment */
	time_t	 tv_sec;	/* seconds after epoch when reassembly
				 * of this packet started */
	uint8_t *bitmap;	/* bitmap representing already 
				 * received parts of the packet */
	char	*buf;		/* buffer into which the fragment is
				 * reassembled */
	LIST_ENTRY(frag_info) frag_infos;	/* queue(3) */
};

/* structure representing a client */
struct client {
	clientid_t	 clid;	/* client ID */
	int		 used;	/* client active or not */
	long		 reqid; /* request ID used for the client ID request */
	struct plugin	*pl;
	/* tunnel device */
	char		 tun_dev[16];
	int		 tun_fd;
	struct event	 tun_ev;
	/* ping */
	int		 ping_counter;
	uint8_t		 echo_seq;
	/* fragmentation */
	struct frag_hdr	 frag_hdr;
	char		 frag_data[MTU+sizeof(struct frag_hdr)];
	char		*frag_datap;
	int		 frag_data_len;
	uint16_t	 frag_id;/* id for the next packet to be fragmented */
	/* fragment reassembly information */
	LIST_HEAD(frag_infos_head, frag_info) frag_infos;
};
/* TODO: randomize initial value of frag_id */
/* TODO: init: frag_datap, frag_data_len, frag_id, echo_seq, client id, */
/*             frag_info_list */


/* FUNCTION HEADERS */
static int process_data_from_tun(struct client *cl, struct plugin *pl,
	       uint8_t dispatch, char *data, int len);
static int send_next_frag();
static void cleanup();
static void send_plugin_select(struct client *cl);


/* GLOBAL VARIABLES */
int server = 0;	/* are we a server or a client? */
TAILQ_HEAD(plugins_head, plugin) plugins = TAILQ_HEAD_INITIALIZER(plugins);
struct client clients[MAXCLIENTS]; /* client is using clients[0] for itself */
static uint8_t myclid = 0; /* client ID of this client (client only) */

/* server host */
struct addrinfo *ai;
char *host;
char *port = "12345";

static struct event timer_ev;

/* dump data in hex format */
static void
dump_data(char *data, int len)
{
	int i;
	printf("(%d bytes) ", len);
	for(i = 0; i < len; i++)
		printf("%02hhx ", *(data+i));
	printf("\n");
}

/*
 * not used yet - useful out, i.e., which plugins work should be
 * produced with this function
 */
int
mylog(const char *fmt, ...)
{
	int out;
	va_list ap;
	
	va_start(ap, fmt);
	out = vprintf(fmt, ap);
	va_end(ap);
	
	return out;
}

/*
 * debugging output - will be removed
 */
int
debug(const char *fmt, ...)
{
	int out;
	va_list ap;
	
	va_start(ap, fmt);
	out = vprintf(fmt, ap);
	va_end(ap);
	
	return out;
}

/* show and execute a command */
int
ssystem(const char *fmt, ...)
{
  char cmd[128];
  va_list ap;
  va_start(ap, fmt);
  vsnprintf(cmd, sizeof(cmd), fmt, ap);
  va_end(ap);
  fprintf(stderr, "%s\n", cmd);
  return system(cmd);
}

/* lookup the client by ID */
static struct client *
lookup_clid(clientid_t clid)
{
	if (server)
		return &clients[clid];
	else
		return clients;
}

/*
 * change the plugin a client is using - closes the connection in the
 * old plugin and adds/deletes the tunnel event depending on the
 * plugin type (polling or direct)
 */
static void
set_client_pl(struct client *cl, struct plugin* pl)
{
	if (cl->pl == pl)
		return;

	if (cl->pl != NULL) {
		cl->pl->conn_close(pl, cl->clid);
		if  (cl->tun_fd != -1)
			event_del(&cl->tun_ev);
	}
	
	cl->pl = pl;
	cl->ping_counter = 0;

	if (pl != NULL && cl->tun_fd != -1) {
		if (!server) /* client */
			send_plugin_select(cl);

		if (cl->pl->is_ready_to_send(cl->pl, cl->clid) ==
		    WILL_SEND_IMMEDIATELY)
			event_add(&cl->tun_ev, NULL);
		// else the tun read is timed by the plugin
	}
}

/*
 * entry function if a polling plugin can send more data - if there
 * are pending fragments, send these; otherwise read data from the tun
 * interface and pass it to the daemon
 */
static void 
request_tun_data(struct client *cl)
{
	char packet[PACKETLEN];
	int nread = 0;
	int nwrite = 0;
	
	printf("request_tun_data()\n");
	if (cl->pl == NULL)
		return;
	
	if (cl->frag_datap != NULL) {
		send_next_frag(cl);
		return;
	}
	
	do {
		nread = tun_read(cl->tun_fd, packet, PACKETLEN);
		if (nread > 0) {
			nwrite = process_data_from_tun(cl, cl->pl,
				     DISPATCH_DATA, packet, nread);
		}
		/* continue the loop if we read something
		 * and the plugin has sent it immediately
		 */
		printf("request_tun_data: nread: %d, nwrite: %d\n", nread, nwrite);
	} while (nread > 0 && nwrite == SEND_PKT_SENT);
}

/* write a packet (data) to the the tun interface of a client */
static int
tun_send(struct client *cl, char *data, int len) {
    int n;
    n = tun_write(cl->tun_fd, data, len);
    /*
    if (n != len) {
	fprintf(stderr, "tun_send: tun_write wrote less bytes (%d) than "
		"requested (%d)\n", n, len);
	// report error
    }
    */
    return n;
}

/* handler function for events on the tun device - called by libevent */
void
tun_ev_handler(int fd, short ev_type, void *arg)
{
	struct client *cl = arg;
	printf("tun_ev_handler(): ev_type: 0x%x\n", ev_type);
	/*
	 * Only read from the tun device can send the data immediately.
	 * Otherwise, we will be notified via plugin_report() when data
	 * can be sent.
	 */
	if (cl->pl != NULL)
		if (cl->pl->is_ready_to_send(cl->pl, cl->clid) ==
		    WILL_SEND_IMMEDIATELY)
			request_tun_data(cl);
}

/* initialize a tun device */
int
tun_init(struct client *cl)
{
	int tun_flags;
	
	/* create the tunnel device */
	sprintf(cl->tun_dev, "tun%i", cl->clid);
	cl->tun_fd = tun_open(cl->tun_dev);
	if (cl->tun_fd < 1) {
		printf("Could not create tunnel device\n");
		return 1;
	}
	/* non-blocking i/o */
	tun_flags = fcntl(cl->tun_fd, F_GETFL, 0);
	if (tun_flags == -1)
		errx(EX_OSFILE, "Failed to get flags from the tun device\n");
	fcntl(cl->tun_fd, F_SETFL, tun_flags|O_NONBLOCK);

 	event_set(&cl->tun_ev, cl->tun_fd, EV_PERSIST | EV_READ,
 		  tun_ev_handler, cl);

	printf("Created tunnel device: %s\n", cl->tun_dev);

	/* configure the tun interface */
	char cmd[123];
	if (server) {
#ifdef __FreeBSD__
		sprintf(cmd,
		    "ifconfig %s mtu 1400 192.168.0.%d 192.168.0.%d",
		    cl->tun_dev, cl->clid * 2, cl->clid * 2 + 1);
		ssystem(cmd);
#else
		warnx("Please configure %s yourself...\n", cl->tun_dev);
#endif
/* #ifdef __linux */
/* 		ssystem("ifconfig tun0 mtu 1400 192.168.0.1"); */
/* 		ssystem("route add 192.168.0.2 tun0"); */
/* #endif */
		
	} else {
#ifdef __FreeBSD__
		sprintf(cmd,
		    "ifconfig %s mtu 1400 192.168.0.%d 192.168.0.%d",
		    cl->tun_dev, cl->clid * 2 + 1, cl->clid * 2);
		ssystem(cmd);
#else
		warnx("Please configure %s yourself...\n", cl->tun_dev);
#endif
/* #ifdef __linux */
/* 		ssystem("ifconfig tun0 mtu 1400 192.168.0.2"); */
/* 		ssystem("route add 192.168.0.1 tun0"); */
/* #endif */
	}
	
	if (cl->pl != NULL) {
		if (cl->pl->is_ready_to_send(cl->pl, cl->clid) ==
		    WILL_SEND_IMMEDIATELY)
			event_add(&cl->tun_ev, NULL);
		/* else the tun read is timed by the plugin */
	}

	return 0;
}

/* deinitialize a tun device */
static void
tun_deinit(struct client *cl)
{
	if (cl->tun_fd != -1) {
		event_del(&cl->tun_ev);
		tun_close(cl->tun_fd, cl->tun_dev);
	}
}

/*
 * load a plugin from file and register it
 * returns a pointer to the plugin on success or NULL on failure
 */
static struct plugin *
load_plugin(char *path) {
	struct plugin *pl;
	int (*plugin_register)(struct plugin*);
	void *handle;
	const char *error;

	pl = malloc(sizeof(struct plugin));
	if (!pl) {
		warnx("failed to malloc plugin: out of memory!\n");
		return NULL;
	}

	/* open the library */
	handle = dlopen (path, RTLD_NOW);
	if (!handle) {
		fputs(dlerror(), stderr);
		return NULL;
	}
	/* load the plugin */
	plugin_register = dlsym(handle, "plugin_register");
	if ((error = dlerror()) != NULL)
		goto error;
	/* prepare the plugin for registration */
	plugin_register(pl);
	
	/* add plugin to the list of loaded plugins */
	TAILQ_INSERT_TAIL(&plugins, pl, plugins);
	
	debug("successfully loaded plugin %s\n", pl->name);
	return (pl);

 error:
	fputs(error, stderr);
	free(pl);
	return (NULL);
}

/* initialize a client */
static struct client *
client_init(clientid_t clid)
{
	struct client *cl = lookup_clid(clid);
	
	if (cl->used) {
		warnx("client_init: client ID already exists\n");
		return NULL;
	}
	
	memset(cl, 0, sizeof(*cl));
	cl->clid = clid;
	cl->tun_fd = -1;
	cl->pl = NULL;
	cl->used = 1;

	return cl;
}

/* deinitialize a client datastructure, the corresponding tun device,... */
static void
client_deinit(struct client *cl)
{
	struct frag_info *np, *np_temp;

	/* close the (plugin) connection */
	if (cl->pl)
		cl->pl->conn_close(cl->pl, cl->clid);
	/* deinitialize the tun device */
	tun_deinit(cl);

    /* deallocate fragment reassembly information */
	LIST_FOREACH_SAFE(np, &cl->frag_infos, frag_infos, np_temp) {
		LIST_REMOVE(np, frag_infos);
		free(np->bitmap);
		free(np->buf);
		free(np);
	}
    
	cl->used = 0;
}

/*
 * send a message to the server notifying it of plugin failover
 */
static void
send_plugin_select(struct client *cl)
{
	int nwrite = 0;
	char data[10];
	char *datap = data;
	int len=0;
	int consumed;


	if (cl->pl != NULL) {
		/* client prepends the client ID */
		if (!server) {
			*datap = cl->clid;
			datap++;
			len++;
		}
		
		/* dispatch */
		*datap = DISPATCH_PLUGIN_SELECT;
		datap++;
		len++;

		nwrite =  cl->pl->send(cl->pl, cl->clid,
				      data, len, NORMAL_DATA, &consumed);
		debug("send_plugin_select(): nwrite: 0x%x\n", nwrite);
	}
}
/*
 * Send a tunnel probe - echo request. Note that this is different
 * from ICMP echo.
 */
static void
send_echo_request(struct client *cl)
{
	int nwrite = 0;
	char data[10];
	char *datap = data;
	int len=0;
	int consumed;

	cl->ping_counter++;

	if (cl->pl != NULL) {
		/* client prepends the client ID */
		if (!server) {
			*datap = cl->clid;
			datap++;
			len++;
		}
		
		/* dispatch */
		*datap = DISPATCH_ECHO_REQUEST;
		datap++;
		len++;

		/* ping sequence number */
		*datap = cl->echo_seq++;
		datap++;
		len++;
		
		nwrite =  cl->pl->send(cl->pl, cl->clid,
				      data, len, NORMAL_DATA, &consumed);
		debug("send_echo_request(): nwrite: 0x%x\n", nwrite);
	}
}

/*
 * send a request for a client ID - used only by the client
 */
static void
send_id_request()
{
	struct client *cl = clients;
	int nwrite = 0;
	char data[10];
	char *pdata = data;
	int consumed;

	if (cl->pl != NULL) {
		/* client ID */
		*pdata = 0;
		pdata++;
		
		/* dispatch */
		*pdata = DISPATCH_ID_REQUEST;
		pdata++;

		/* request ID */
		memcpy(pdata, &cl->reqid, sizeof(cl->reqid));
		pdata += sizeof(cl->reqid);
		
		nwrite =  cl->pl->send(cl->pl, cl->clid,
				      data, pdata - data, NORMAL_DATA,
				      &consumed);
		debug("send_id_request(): nwrite: 0x%x\n", nwrite);
	}
}



/* handler function for the libevent timer event */
static void 
timer_ev_handler(int fd, short ev_type, void *arg)
{
	struct timeval tv;
	struct frag_info *np, *np_temp;
	struct client *cl;
	int i;

	/* check if too many ping requests have not failed */
	for(i = 0; i < MAXCLIENTS; i++) {
		cl = &clients[i];
		if (cl->used == 0)
			continue;
		
		printf("ping_counter: %d\n", cl->ping_counter);
		if (cl->ping_counter > PING_FAIL_COMPLETE &&
		    server)
			client_deinit(cl);
		else if (cl->pl) {
			if (cl->ping_counter > PING_FAIL_PLUGIN)
				plugin_report(cl->pl, cl->clid,
				    REPORT_ERROR_PING);
			send_echo_request(cl);
		}
	}
	
	/* fragment reassembly timeout */
	gettimeofday(&tv, NULL);
	for(i = 0; i < MAXCLIENTS; i++) {
		cl = &clients[i];
		if (cl->used == 0)
			continue;
		
		LIST_FOREACH_SAFE(np, &cl->frag_infos, frag_infos, np_temp) {
			if (tv.tv_sec - np->tv_sec > FRAG_TIMEOUT) {
				LIST_REMOVE(np, frag_infos);
				free(np->bitmap);
				free(np->buf);
				free(np);
			}
		}
	}
    
	/* register a timer event again */
	tv.tv_sec=PING_INTERVAL;
	tv.tv_usec=0;
	evtimer_set(&timer_ev, timer_ev_handler, NULL);
	evtimer_add(&timer_ev, &tv);
}

/*
 * Pass data received by the plugin to the daemon. This function is
 * called from plugins. Before calling plugin_send, it should call
 * plugin_conn_map().
 *
 * clid and conn_flag will be removed from the function signature as
 * plugin_conn_map() is used now.
 */
void
process_data_from_plugin(struct plugin *pl, char *data, int len,
    uint8_t *clid, int *conn_flag)
{
	u_int8_t dispatch;
	int i;
	int nwrite;
	int consumed;
	struct client *cl = NULL;

	/* fragment reassembly */
	struct frag_hdr *frag_hdr = NULL;
	struct frag_info *p;	/* pointer to frag info about the
				 * processed fragment */
	struct frag_info *np;
	struct timeval tv;
	int dgram_reassembled;

/* 	debug("data from plugin: "); */
/* 	dump_data(data, len); */

	*conn_flag = CONN_DISCARD;

	if (len <= 0)
		return;

	/* process the client ID */
	if (server) {
		/* payload from the client starts with the client ID */
		*clid = *data;
		data++;
		len--;
	} else { /* client */
		/* payload from the server does NOT contain the client ID */
		*clid = 0;
	}
	cl = lookup_clid(*clid);

	/* process the dispatch */
	dispatch = *data;
	debug("clid: %hhd, dispatch: 0x%02hhx\n", *clid, dispatch);
	switch (dispatch) {
	case DISPATCH_DATA:
		debug("process_data_from_plugin(): DATA\n");
		/* only associated clients can send DATA to the server */
		if (server && *clid == 0) {
			*conn_flag = CONN_DISCARD;
			pl->conn_map(pl, *clid, *conn_flag);
			return;
		}

		*conn_flag = CONN_PERM;
		pl->conn_map(pl, *clid, CONN_PERM);
		
		/* update the current plugin for this client */
		set_client_pl(cl, pl);
		
		/* pass data to the tun device */
		tun_send(cl, data+1, len-1);

		break;

	case DISPATCH_PLUGIN_SELECT:
		debug("process_data_from_plugin(): PLUGIN_SELECT\n");
		/* only associated clients can send DATA to the server */
		if (server && *clid == 0) {
			*conn_flag = CONN_DISCARD;
			pl->conn_map(pl, *clid, *conn_flag);
			return;
		}

		*conn_flag = CONN_PERM;
		pl->conn_map(pl, *clid, CONN_PERM);
		
		/* update the current plugin for this client */
		set_client_pl(cl, pl);
		
		break;

	case DISPATCH_FRAG: /* fragment reassembly */
		if (len <= sizeof(*frag_hdr)) {
			plugin_report(pl, *clid, REPORT_ERROR_RECEIVE);
			return;
		}

		frag_hdr = (struct frag_hdr*) data;
		data += sizeof(struct frag_hdr);
		len -= sizeof(struct frag_hdr);
		
		debug("got a frag header: id %d, size %d, off %d, len %d\n",
		    frag_hdr->id, frag_hdr->size, frag_hdr->offset, len);

		/* check if reassembly for this fragment has already started */
		p = NULL;
		LIST_FOREACH(np, &cl->frag_infos, frag_infos) {
			if (np->id == frag_hdr->id &&
			    np->size == frag_hdr->size) {
				/* found in list */
				debug("found frag info in list\n");
				p = np;
				break;
			}
		}

		/* fragment info not found in list, start a new reassembly */
		if (p == NULL) {
			fprintf(stderr, "frag info NOT found in list\n");
			/* allocate memory */
			p = malloc(sizeof(struct frag_info));
			if (!p) {
				warnx("process_data_from_plugin - "
				    "fragment reassembly: out of memory\n");
				return;
			}
			memset(p, 0, sizeof(*p));
			p->buf = malloc(frag_hdr->size);
			if (!p->buf) {
				free(p);
				warnx("process_data_from_plugin - "
				    "fragment reassembly: out of memory\n");
				return;
			}
			p->bitmap = malloc(frag_hdr->size / 8 + 1);
			if (!p->bitmap) {
				free(p->buf);
				free(p);
				warnx("process_data_from_plugin - "
				    "fragment reassembly: out of memory\n");
				return;
			}
			memset(p->bitmap, 0, frag_hdr->size / 8 + 1);

			/* collect information about the fragments */
			gettimeofday(&tv, NULL);
			p->id = frag_hdr->id;
			p->size = frag_hdr->size;
			memcpy(&p->tv_sec, &tv.tv_sec, sizeof(tv.tv_sec));
			LIST_INSERT_HEAD(&cl->frag_infos, p, frag_infos);
		}
	
		if (frag_hdr->offset + len > p->size) {
			debug("fragment payload outside of packet payload\n");
			return;
		}
		/* copy the data */
		memcpy(p->buf + frag_hdr->offset, data, len);
		/* update the bitmap
		 * 
		 * We ignore fragment overlaps as these should
		 * be caught by the upper layer
		 * checksum. Actually, they should not happen
		 * at all.
		 */
		for (i = frag_hdr->offset; i < frag_hdr->offset+len; i++)
			p->bitmap[i/8] |= (0x1 << (i%8));
		
		/* 
		 * Fragment was processed without errors, so it was
		 * valid traffic and the plugin used by this client
		 * should be updated.
		 */
		pl->conn_map(pl, *clid, CONN_PERM);
		set_client_pl(cl, pl);
	
		/* check if the complete packet has been reassembled */
		dgram_reassembled = 1;
		/* examine the bitmap */
		for(i = 0; i < p->size / 8 && dgram_reassembled; i++) {
			if (p->bitmap[i] != 0xff) {
				dgram_reassembled = 0;
			}
		}
		for(i = 0; i < p->size % 8 && dgram_reassembled; i++) {
			if ((p->bitmap[p->size / 8] & (1 << i)) == 0) {
				dgram_reassembled = 0;
			}
		}
		
		/* packet completely reassembled */
		if (dgram_reassembled != 0) {
			debug("frag reassembly: packet complete\n");
			
			/* pass the reassembled packet to the tun device */
			tun_send(cl, p->buf, p->size);
	    
			/* fragment info for this packet no longer needed */
			LIST_REMOVE(p, frag_infos);
			free(p->buf);
			free(p->bitmap);
			free(p);
		}

		break;

	case DISPATCH_ECHO_REQUEST:
		debug("process_data_from_plugin(): ECHO_REQUEST\n");
		debug("got echo request (plugin: %s)\n", pl->name);
		debug("seq: %u\n", *(data+1));
		/* pings are qualify for temporary connection status only */
		*conn_flag = CONN_TEMP;
		pl->conn_map(pl, *clid, CONN_TEMP);
		
		/* generate a reply to the echo request */
		*data = (u_int8_t)DISPATCH_ECHO_REPLY;

		/* client only - prepend the client ID */
		char *ndata;
		if (!server) { /* client */
			/* 
			 * payload from the client should contain the
			 * client ID so prepend it
			 */
			ndata = malloc(len + 1);
			if (!ndata) {
				warnx("out of memory - cannot generate an "
				      " echo reply\n");
				return;
			}
			memcpy(ndata+1, data, len);
			data = ndata;
			*data = myclid;
		}
		
		nwrite = pl->send(pl, 0, data, len, URGENT_DATA, &consumed);
		debug("sending reply, returned 0x%x\n", nwrite);

		if (server) {
			if (*clid != 0)
				set_client_pl(cl, pl);
		} else /* client */
			free(data);

		break;

	case DISPATCH_ECHO_REPLY:
		debug("process_data_from_plugin(): ECHO_REPLY\n");
		debug("got echo reply (plugin: %s)\n", pl->name);
		
		/* pings are qualify for temporary connection status only */
		pl->conn_map(pl, *clid, CONN_TEMP);
		*conn_flag = CONN_TEMP;
		
		/* update the ping counter */
		cl->ping_counter = 0;
		
		/* 
		 * Receiving a ping reply indicates a working plugin.
		 * This is a good point to update the currently used
		 * plugin if none is set and to request a client ID if
		 * not associated already.
		 */
		if (!server) { /* client */
			if(cl->pl == NULL)
				set_client_pl(0, pl);
			/* request a client ID from the server */
			if (myclid == 0)
				send_id_request();
		}
		
		break;
		
	case DISPATCH_ID_REQUEST:
		debug("process_data_from_plugin(): ID_REQUEST\n");
		/* only relevant to the server */
		if (!server) {
			pl->conn_map(pl, *clid, CONN_DISCARD);
			return;
		}

		/* the client ID in a request has to be zero */
		if (*clid != 0) {
			pl->conn_map(pl, *clid, CONN_DISCARD);
			return;
		}
		
		//TODO reqid = *((long*)(data+1));

		/* assign a new client ID */
		for (i = 1; i < MAXCLIENTS && clients[i].used != 0; i++);
		if (i == MAXCLIENTS) {
			*conn_flag = CONN_DISCARD;
			warnx("Cannot accept any more clients\n");
			return;
		}

		/* found an available ID, associate it with the client */
		*clid = i;
		*conn_flag = CONN_PERM;
		pl->conn_map(pl, *clid, CONN_PERM);
		cl = client_init(*clid);
		cl->clid = *clid;
		tun_init(cl);
		set_client_pl(cl, pl);

		/* reply to the client offering the ID */
		*(data) = DISPATCH_ID_OFFER;
		*(data+1) = *clid;
		nwrite = pl->send(pl, *clid, data, 2, URGENT_DATA, &consumed);
		debug("got ID request (plugin: %s)\n", pl->name);
		debug("sending an offer, returned 0x%x\n", nwrite);

		break;
		
	case DISPATCH_ID_OFFER:
		debug("process_data_from_plugin(): ID_OFFER\n");
		/* only relevant to the client */
		if (server) {
			pl->conn_map(pl, *clid, CONN_DISCARD);
			return;
		}
		
		if (myclid != 0) {  /* ID already assigned */
			//pl->conn_map(pl, *clid, CONN_DISCARD);
			return;
		}

		/* parse the reply for the client ID */
		myclid = *(data+1);
		cl->clid = myclid;
		debug("got an ID offer (%d)\n", myclid);
		
		/* 
		 * We can start tunneling after a successful
		 * association with a server so initialize the tun
		 * device.
		 */
		tun_init(clients);
		
		break;
		
	case DISPATCH_PLUGIN_DATA:
		/* plugin-to-plugin communication */
		debug("process_data_from_plugin(): PLUGIN_DATA\n");
		pl->conn_map(pl, *clid, CONN_TEMP);

		/* pass the payload to the plugin */
		pl->custom_receive(pl, 0, data + 1, len - 1);
		
		break;

	default:
		debug("unknown dispatch 0x%X in data from plugin "
		      "%s\n",	dispatch, pl->name);
		pl->conn_map(pl, *clid, CONN_DISCARD);
	}
}

/*
 * Pass data received from the tun interface to the daemon.
 */
static int
process_data_from_tun(struct client *cl, struct plugin *pl, uint8_t dispatch,
    char *data, int len)
{
	int consumed;
	char ldata[MTU+3];

	if (pl == NULL)
		pl = cl->pl;

	if (pl == NULL) {
		debug("no plugin connected yet, discarding tun data\n");
		plugin_report(pl, cl->clid, REPORT_BOOTSTRAP);
		return SEND_ERROR;
	}

	debug("process_data_from_tun: len: %d, pl->mtu: %d\n",
	    len, pl->mtu);
	/* no need to add the fragmentation header */
	if (len < pl->mtu || dispatch == DISPATCH_PLUGIN_DATA) {
		if (server) {
			*(ldata) = dispatch;
			memcpy(ldata + 1, data, min(sizeof(ldata)-1, len));
			return (pl->send(pl, cl->clid, ldata,
			    min(sizeof(ldata), len + 1), NORMAL_DATA,
			    &consumed));
		} else { /* client */
			*ldata = myclid;
			*(ldata + 1) = dispatch;
			memcpy(ldata+2, data, min(sizeof(ldata)-2, len));
			return (pl->send(pl, myclid, ldata,
			    min(sizeof(ldata), len + 2), NORMAL_DATA,
			    &consumed));
		}
	/* add the fragmentation header */
	} else {
		/* prepare the frag header */
		cl->frag_hdr.dispatch = DISPATCH_FRAG;
		cl->frag_hdr.id = cl->frag_id;
		cl->frag_hdr.size = len;
		cl->frag_hdr.offset = 0;
		
		/* increment the fragment ID */
		if (cl->frag_id == 65535)
			cl->frag_id = 0;
		else
			cl->frag_id++;

		cl->frag_data_len = 0;
		cl->frag_datap = cl->frag_data;
		
		/* copy the payload */
		if (server) {
			memcpy(cl->frag_datap + sizeof(cl->frag_hdr), data,
			    min(sizeof(ldata) - sizeof(cl->frag_hdr), len));
			cl->frag_data_len = len + sizeof(cl->frag_hdr);
		} else {
			memcpy(cl->frag_datap + sizeof(cl->frag_hdr) + 1, data,
			    min(sizeof(ldata) - sizeof(cl->frag_hdr) - 1,len));
			cl->frag_data_len = len + sizeof(cl->frag_hdr) + 1;
		}
		
		return send_next_frag(cl);
	}
}

/* plugin can request sending data, used for plugin-to=plugin communication */
int
plugin_custom_send(struct plugin *pl, uint8_t clid,
    char *data, int len)
{
	return process_data_from_tun(&clients[clid], pl, DISPATCH_PLUGIN_DATA,
		   data, len);
}

/* send the next fragment */
static int
send_next_frag(struct client *cl)
{
	int nwrite = SEND_PKT_SENT;
	int consumed;
	int n;

	debug("send_next_frag: frag_hdr.size: %u\n", cl->frag_hdr.size);
	if (server) {
		while(cl->frag_data_len > sizeof(cl->frag_hdr) &&
		      nwrite == SEND_PKT_SENT) {
			debug("send_next_frag: sending frag "
			       "offset %d, send_len %d, frag_data_len %d\n",
			       cl->frag_hdr.offset,
			       min(cl->pl->mtu, cl->frag_data_len),
			       cl->frag_data_len);
			/*
			 * prepend the frag dispatch and 
			 * header before the payload
			 */
			memcpy(cl->frag_datap, &cl->frag_hdr,
			       sizeof(cl->frag_hdr));
			
			/* send it */
			//TODO: maybe we should check how much data
			//      was actually sent
			if (cl->pl->mtu > 0)
				n = min(cl->pl->mtu, cl->frag_data_len);
			else
				n = cl->frag_data_len;
			nwrite = cl->pl->send(cl->pl, cl->clid,
			    cl->frag_datap, n, NORMAL_DATA, &consumed);
			debug("send_next_frag: consumed: %d\n", consumed);
			switch (nwrite) {
			case SEND_PKT_SENT:
			case SEND_PKT_QUEUED:
				consumed -= sizeof(cl->frag_hdr);
				cl->frag_datap += consumed;
				cl->frag_hdr.offset += consumed;
				cl->frag_data_len -= consumed;
				break;
			default:
				/*plugin_report(current_pl, REPORT_ERROR_SEND);*/
				return nwrite;
			}
		}
		if (cl->frag_data_len <= sizeof(cl->frag_hdr)) {
			cl->frag_datap = NULL;
			cl->frag_data_len = 0;
		}

	} else { /* client */
		while(cl->frag_data_len > sizeof(cl->frag_hdr) + 1 &&
		      nwrite == SEND_PKT_SENT) {
			debug("send_next_frag: sending frag "
			       "offset %d, send_len %d, frag_data_len %d\n",
			       cl->frag_hdr.offset,
			       min(cl->pl->mtu, cl->frag_data_len),
			       cl->frag_data_len);
			/*
			 * prepend the frag dispatch and 
			 * header before the payload
			 */
			*(cl->frag_datap) = myclid;
			memcpy(cl->frag_datap + 1, &cl->frag_hdr,
			       sizeof(cl->frag_hdr));
			
			/* send it */
			if (cl->pl->mtu > 0)
				n = min(cl->pl->mtu, cl->frag_data_len);
			else
				n = cl->frag_data_len;
			nwrite = cl->pl->send(cl->pl, cl->clid,
			    cl->frag_datap, n, NORMAL_DATA, &consumed);
			debug("send_next_frag: consumed: %d\n", consumed);
			switch (nwrite) {
			case SEND_PKT_SENT:
			case SEND_PKT_QUEUED:
				consumed -= sizeof(cl->frag_hdr) + 1;
				cl->frag_datap += consumed;
				cl->frag_hdr.offset += consumed;
				cl->frag_data_len -= consumed;
				break;
			default:
				/*plugin_report(current_pl, REPORT_ERROR_SEND);*/
				return nwrite;
			}
		}
		if (cl->frag_data_len <= sizeof(cl->frag_hdr) + 1) {
			cl->frag_datap = NULL;
			cl->frag_data_len = 0;
		}		
	}
	
	return nwrite;
}

/* 
 * entry function for plugins to report various things to the daemon,
 * such as errors, availability to send mroe data (all queued data has
 * been sent),...
 */
void
plugin_report(struct plugin *pl, clientid_t clid, int err)
{
	struct client *cl;

	cl = lookup_clid(clid);

	if (pl)
		printf("plugin_report(%s, 0x%x)\n", pl->name, err);
	else
		printf("plugin_report(NULL plugin, 0x%x)\n", err);
	
	switch (err) {
	case REPORT_READY_TO_SEND:
		/* plugin is ready to send the next packet */
		if (cl != NULL)
			request_tun_data(cl);
		return;

	/* not used yet */
/* 	case REPORT_INIT_SUCCESS:	/\* initialization succeeded *\/ */
/* 	case REPORT_INIT_FAIL:		/\* initialization failed *\/ */

	/* errors */
	case REPORT_ERROR_PING:		/* too many probes have failed */
	case REPORT_ERROR_SEND:		/* problem with sending */
	case REPORT_ERROR_RECEIVE:	/* problem with receiving */
		pl->conn_close(pl, clid);

		/* FALLTHROUGH */

	case REPORT_BOOTSTRAP:		/* no plugin initialized yet */
	default:
		if (server) {
			if (pl) {
				warnx("plugin %s failed\n", pl->name);
			}
		} else { /* client */
			if (pl) {
				/* deinitialize the broken plugin */
				mylog("plugin failed: %s\n", pl->name);
				pl->deinitialize(pl);
				
				/* move to next plugin */
				pl = TAILQ_NEXT(pl, plugins);
			} else
				pl = TAILQ_FIRST(&plugins);
			
			for (; pl; pl = TAILQ_NEXT(pl, plugins)) {
				/* try to initialize plugin */
				debug("initalizing plugin: %s...\n",
				      pl->name);
				if (((pl->initialize)(pl, host, pl->name+4))
				    == 0) {
					mylog("initialized plugin %s\n",
					    pl->name);
					set_client_pl(cl, pl);
					cl->ping_counter = 0;
					send_echo_request(cl);
					return;
				} else {
					mylog("plugin %s failed to "
					    "initialize\n", pl->name);
				}
			}
			mylog("all plugins failed, giving up\n");
			cleanup();
		}
	}
}

/* garbage collection before exiting */
void
cleanup()
{
	struct plugin *pl;

	event_del(&timer_ev);


	/* deinitialize all clients */
	int i;
	for (i = 0; i < MAXCLIENTS; i++)
		client_deinit(&clients[i]);

	/* deinitialize all plugins */
	while (!TAILQ_EMPTY(&plugins)) {
		pl = TAILQ_FIRST(&plugins);
		TAILQ_REMOVE(&plugins, pl, plugins);
		pl->deinitialize(pl);
		free(pl);
	}

	if (!server)
		freeaddrinfo(ai);

	exit(EX_OK);
}

/* signal handler */
void
sigcb(int sig)
{
    cleanup();
}

void
usage() {
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "server: mtund -s \n");
    fprintf(stderr, "client: mtund -c host\n");
}

/* 
 * Magic Tunnel Daemon - tunnels IP traffic over UDP, TCP, ICMP,
 * DNS,... The various encapsulation are implemented as run-time
 * loadable plugins. The daemon automatically probes for a working
 * encapsulation, tunnels ovber it and can failover to a different
 * encapsulation if it fails.
 */
int
main(int argc, char **argv) {
	struct plugin *pl;
	struct addrinfo hints;
	int ngai;
	char opt;

	/* argument parsing */
	while((opt=getopt(argc, argv, "sc")) != EOF) {
		switch(opt) {
		case 's':
			server = 1;
			break;
		case 'c':
			server = 0;
			break;
		default :
			fprintf(stderr, "unknown parameter\n");
			exit(EXIT_FAILURE);
			break;
		}
	}
	argc -= optind;
	argv += optind;
	
	if (server) {
		if (argc != 0) {
			usage();
			exit(EXIT_FAILURE);
		}
	} else {
		if (argc == 1) {
			host = argv[0];
		} else {
			usage();
			exit(EXIT_FAILURE);
		}
	}

	/* Initalize libevent */
	event_init();

	/* install signal handlers */
	signal(SIGHUP, sigcb);
	signal(SIGINT, sigcb);
	signal(SIGTERM, sigcb);
	
	/* load plugins */ 
/* 	if (server) { */
/* 		pl = load_plugin("./plugin_udp_catchall.so"); */
/* 		pl->name = "udp_catchall"; */
/* 	} else { /\* client *\/ */
/* 		pl = load_plugin("./plugin_udp.so"); */
/* 		pl->name = "udp_1234"; */
/* 		pl = load_plugin("./plugin_udp.so"); */
/* 		pl->name = "udp_1235"; */
/* /\* 		pl = load_plugin("./plugin_udp.so"); *\/ */
/* /\* 		pl->name = "udp_53"; *\/ */
/* 	} */
/* 	pl = load_plugin("./plugin_tcp.so"); */
/* 	pl->name = "tcp_1234"; */
	pl = load_plugin("./plugin_icmp.so");
	pl->name = "icmp";
	pl = load_plugin("./plugin_dns/plugin_dns.so");
	pl->name = "dns_53";

	if (server) {
		/* initialize all plugins */
		TAILQ_FOREACH(pl, &plugins, plugins) {
			//(void) (pl->initialize)(server, host, port);
			(void) (pl->initialize)(pl, host, pl->name+4);
			// we should unload plugins which fail to initialize
		}
	} else { /* client */
		/* DNS lookup */
		memset(&hints, 0, sizeof(hints));
		hints.ai_family = AF_INET;
		if ((ngai = getaddrinfo(host, NULL, &hints, &ai)) != 0)
			errx(EX_NOHOST, "getaddrinfo: %s\n",
			   gai_strerror(ngai));
		
		/* setup the client */
		client_init(0);
		srandomdev();
		clients[0].reqid = random();
		clients[0].frag_id = random();
		
		/* initialize a working plugin */
		plugin_report(NULL, 0, REPORT_BOOTSTRAP);
	}
	
	/* kick off the timer event, it will re-register itself */
	timer_ev_handler(-1, 0, NULL);
	
	/* wait for events and process them */
	event_dispatch();
	
	cleanup();
	return 1;
}
