/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _TUNNELD_H
#define _TUNNELD_H

#include <sys/types.h>
#include <sys/queue.h>

#define PACKETLEN 1400   /* encapsulated data size */
#define SOCKET_TIMEOUT 5 /* timeout for socket operations */
#define FRAG_TIMEOUT 60  /* encapsulated data size */

#define MAXCLIENTS 16

typedef uint8_t clientid_t;	/* client ID */
typedef void* reqid_t;		/* request identifier */
typedef uint8_t dispatch_t;


/* structure representing a plugin in the main tunnel daemon */
struct plugin {
	int (*initialize)(struct plugin*, char*, char*);
	void (*deinitialize)(struct plugin*);
	void (*receive)(int fd, short ev_type, void *arg); /* select fired on some fd - check for data */
	int (*send)(struct plugin*, uint8_t, char*, int, int, int*);
	int (*is_ready_to_send)(struct plugin*, uint8_t);
	void (*conn_close)(struct plugin*, uint8_t);
	void (*conn_map)(struct plugin*, uint8_t, int);
	void (*custom_receive)(struct plugin*, uint8_t, char*, int);
	char *name;
	int mtu;
	void* data;
	TAILQ_ENTRY(plugin) plugins;
};

/* plugin_report() argument values */
enum {
    REPORT_READY_TO_SEND,	/* ready to send the next packet */
    REPORT_INIT_SUCCESS,	/* initialization succeeded */
    REPORT_INIT_FAIL,		/* initialization failed */
    REPORT_BOOTSTRAP,		/* no plugin initialized yet */
    REPORT_ERROR_SEND,		/* problem with sending */
    REPORT_ERROR_RECEIVE,	/* problem with receiving */
    /*REPORT_ERROR_TIMEOUT,*/
    REPORT_ERROR_PING,		/* too many pings have failed,
				 * failover to another plugin */
};

/* plugin_is_ready_to_send() return values */
enum {
    WILL_SEND_IMMEDIATELY,
    WILL_QUEUE,
    QUEUE_FULL
};

/* plugin_send() argument */
enum {
    NORMAL_DATA,    /* normal data */
    URGENT_DATA     /* urgent data */
};

/* plugin_send() return values */
enum {
    SEND_PKT_SENT = 0,     /* packet has been sent,
			    * ready to send the next one
			    */
    SEND_PKT_QUEUED,       /* packet queued, not yet ready
			    * to send the next one
			    */
    SEND_ERROR_MALLOC,     /* malloc failed, packet dropped */
    SEND_ERROR_QUEUE_FULL, /* queue is full, packet dropped */
    SEND_ERROR             /* sending the packet failed */
};

/* dispatch type - prepended before the payload */
enum {
    DISPATCH_ECHO_REQUEST,
    DISPATCH_ECHO_REPLY,
    DISPATCH_PLUGIN_SELECT,
    DISPATCH_FRAG,
    DISPATCH_ID_REQUEST,	/* client requesting an ID */
    DISPATCH_ID_OFFER,		/* daemon offering an ID to the client */
    DISPATCH_PLUGIN_DATA,	/* plugin to plugin communication */
    DISPATCH_DATA         = 0x42,
};

/* 
 * conn_flag argument of process_data_from_plugin()
 * flags indicating whether connection should be kept by plugin
 */
enum {
	CONN_DISCARD,	/* data was garbage, connection can be discarded */
	CONN_TEMP,	/* connection from an unassociated client, store
			 * connection metadata temporarily (until timeout) */
	CONN_PERM	/* connection from an associated client, store
			 * connection metadata, override previous metadata */
};

extern int server; /* are we a server (0) or a client (1)? */

/*
 * Pass data received by the plugin to the daemon.
 */
void process_data_from_plugin(struct plugin *pl, char *data, int len,
			      uint8_t *clid, int *conn_flag);

/*
 * Report a problem in the plugin to the daemon.
 */
void plugin_report(struct plugin *pl, uint8_t clid, int err);

/*
 * plugin can request sending data, used for plugin-to=plugin communication
 */
int plugin_custom_send(struct plugin *pl, uint8_t clid, char *data, int len);

/*
 * debugging output
 */
int debug(const char *fmt, ...);

#endif
