/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _PLUGIN_H
#define _PLUGIN_H

#include "mtund.h"

/* plugin state */
enum {
    PLUGIN_STATE_UNINITIALIZED,
    PLUGIN_STATE_INITIALIZED,
    PLUGIN_STATE_CONNECTED,
    PLUGIN_STATE_DISCONNECTED
};

/* 
 * A plugin has to implement all the plugin_* functions. Note that
 * none of the plugin functions should block! At least not
 * indefinitely.
 */

/*
 * Completely fill in pl with function pointers and name of the
 * plugin. No need to initialize yet. Expect plugin_initialize() to be
 * called.
 * Return: 0 on success, -1 on failure
 */
int plugin_register(struct plugin *pl);

/*
 * Clean up pl and memory alocations there.
 */
void plugin_unregister(struct plugin* pl);

/* Initialize the plugin, open a socket for listening, register its
 * file descriptor with register_select_fd() and expect
 * plugin_receive() to be called. Server specifies if we're a server
 * (1) or a client (0).
 *
 * Return: 0 on success, -1 on failure
 *
 * For server, success usually means a listeninging socket has been
 * opened. For a client, success means a socket has been connected and
 * in case of UDP handshake data exchanged.
 */
int plugin_initialize(struct plugin *pl, char *host, char *port);

/*
 * Deinitialize the plugin, close open sockets, unregister them with
 * unregister_select_fd() and free allocated memory.
 */
void plugin_deinitialize(struct plugin *pl);

/*
 * Check with fd_isset() whether your file descriptor has data and if
 * so, process it.
 * Return: number of bytes received and sent out by tun_send(), -1 on error.
 * If plugin was not active, but got active now, i.e. a connection
 * from the client was accepted, return >0 so that it becomes the
 * current_pl!
 */
void plugin_receive(int fd, short ev_type, void *arg);

/*
 * Close connection for given client
 */
void plugin_conn_close(struct plugin *pl, uint8_t clid);

/*
 * Map the most recently receiving connection to a client ID.
 * This function has to be called from process_data_from_tun() before
 * calling plugin_send()!
 */
void plugin_conn_map(struct plugin *pl, uint8_t clid, int conn_flag);

/*
 * Check whether the plugin is ready to send data and it if would send
 * it immediately or queue (return values defined in mtund.h).
 */
int plugin_is_ready_to_send(struct plugin *pl, uint8_t clid);

/*
 * Send the data, using connection corresponding to the client ID
 * (clid) ot the most recently receiving connection (clid 0). Return
 * values are defined in mtund.h. consumed indicates the number of
 * bytes of data sent.
 */
int plugin_send(struct plugin *pl, uint8_t clid,
    char *data, int len, int data_type, int *consumed);

/*
 * Data for the plugin from plugin-to-plugin communication.
 */
void
plugin_custom_receive(struct plugin *pl, uint8_t clid,
		      char *data, int len);

#endif
