/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * DNS encapsulation plugin
 *
 * calls functions from iodine source code to do the dns en/decoding
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <err.h>
#include <errno.h>
#include <event.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "../mtund.h"
#include "../plugin.h"

#include "common.h"
#include "dns.h"
#include "encoding.h"

/*
 * how often should an empty request be sent to the server - This is
 * useful when the server has data to send but the client doesn't.
 */
#define PLUGIN_DNS_KEEP_ALIVE_SEC 0
#define PLUGIN_DNS_KEEP_ALIVE_USEC 500000

#define TOPDOMAIN "matustest.eecs.jacobs-university.de"

struct conn {
	clientid_t clid;
	int		 used;

	/* normal data queue - regular traffic from the tun interface
	 * and ping requests */
	char		*queued_normal_data;
	int		 queued_normal_data_len;
	/* urgent data queue - ping replies */
	char		*queued_urgent_data;
	int		 queued_urgent_data_len;
	/*
	 * Has the client sent data after the last reply from the
	 * server was received?
	 */
	int		 data_sent_after_last_receive; /* client only */
};

struct plugin_dns_data {
	int		 fd; /* DNS socket */
	int		 state;

	/* last request */
	struct sockaddr_storage addr;
	socklen_t	 addrlen;
	struct query	 q;
	struct event	 ev; /* used by libevent to monitor our fd */
	struct event	 timer_ev;
	struct conn	 conns[MAXCLIENTS];
	struct conn	*conn; /* most recently receiving conn */
	char		*topdomain;
};

static void plugin_send_queued_data();
static void plugin_send_empty_request(struct plugin *pl, struct conn *conn);

static int
udp_connect(char *host, char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_DGRAM;

    n = getaddrinfo(host, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(n));
        exit(EXIT_FAILURE);
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }
        if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        fprintf(stderr, "socket or connect: failed for %s port %s\n",
		host, port);
//        exit(EXIT_FAILURE);
                return -1;
    }

    return fd;
}

/*
 * Create a named UDP endpoint. First get the list of potential
 * network layer addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create a UDP
 * endpoint is successful (or no other alternative exists).
 */

static int
udp_open(char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, on = 1;

    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_DGRAM;

    n = getaddrinfo(NULL, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "getaddrinfo failed: %s\n", gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }

        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
        if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        fprintf(stderr, "bind failed for port %s\n", port);
        return -1;
    }

    return fd;
}

/* register a timer event */
static void
register_timer_ev(struct event *ev)
{
    struct timeval tv;
    
    tv.tv_sec=PLUGIN_DNS_KEEP_ALIVE_SEC;
    tv.tv_usec=PLUGIN_DNS_KEEP_ALIVE_USEC;
    evtimer_del(ev);
    evtimer_add(ev, &tv);
}

int
dns_send(struct plugin_dns_data *pldata, struct conn *conn,
		char *data, int datalen)
{
	char buf[PACKETLEN];
	char packet[PACKETLEN];
	size_t len;
	size_t consumed;
	int n;
	char tmpdata[4];
	
	
	/* DNS encoding/decoding seems to be broken for too short payloads */
	if (datalen < 3) {
		memset(tmpdata, 0, sizeof(tmpdata));
		memcpy(tmpdata, data, datalen);
		data = tmpdata;
		datalen=sizeof(tmpdata);
	}

	if (server) {
		len = dns_encode(buf, sizeof(buf), &pldata->q, QR_ANSWER,
				 data, datalen);
		n = sendto(pldata->fd, buf, len, 0,
			   (struct sockaddr*)&pldata->addr, pldata->addrlen);
		return datalen;
	} else { /* client */
		pldata->q.id++;
		pldata->q.type = T_NULL;
	
		consumed = dns_build_hostname(buf, sizeof(buf),
			       data, datalen, pldata->topdomain);
		len = dns_encode(packet, sizeof(packet), &pldata->q, QR_QUERY,
				 buf, strlen(buf));
	
		n = send(pldata->fd, packet, len, 0);
		
		/* the client has to reset the timer for keep-alive requests */
		register_timer_ev(&pldata->timer_ev);
		printf("dns_send(): consumed: %d\n", consumed);
		return consumed;
	}
}

static void
conn_init(struct conn *conn)
{
	conn->used = 0;
	conn->clid = 0;

	conn->queued_normal_data = NULL;
	conn->queued_normal_data_len = 0;
	conn->queued_urgent_data = NULL;
	conn->queued_urgent_data_len = 0;	
}

static struct conn*
conn_by_clid(struct plugin_dns_data *data, clientid_t clid)
{
	if (server)
		if (clid == 0)
			return data->conn;
		else
			return &data->conns[clid];
	else
		return data->conns;
}

int
plugin_register(struct plugin* pl)
{
    pl->name = "dns";
    if (server) 
	    //pl->mtu = 1024;
	    pl->mtu = 512;
    else /* client */
	    pl->mtu = 0; /* always prepend the fragment header */
    pl->initialize = plugin_initialize;
    pl->deinitialize = plugin_deinitialize;
    pl->is_ready_to_send = plugin_is_ready_to_send;
    pl->send = plugin_send;
    pl->receive = plugin_receive;
    pl->conn_close = plugin_conn_close;
    pl->conn_map = plugin_conn_map;
    pl->custom_receive = plugin_custom_receive;

    pl->data = malloc(sizeof(struct plugin_dns_data));
    if (!pl->data) {
	return -1;
    }
    struct plugin_dns_data *data = (struct plugin_dns_data*) pl->data;
    data->fd = -1;
    data->topdomain = TOPDOMAIN;
    data->state = PLUGIN_STATE_UNINITIALIZED;

    return 0;
}

void
plugin_unregister(struct plugin* pl)
{
    if (pl->data) {
	free(pl->data);
	pl->data = NULL;
    }
}

/* handler function for the libevent timer event - client only */
void 
plugin_dns_timer_ev_handler(int fd, short ev_type, void *arg)
{
    struct plugin_dns_data *data = ((struct plugin *)arg)->data;

    plugin_send_empty_request(arg, data->conn);

     /* register a timer event again */
    register_timer_ev(&data->timer_ev);
}

int
plugin_initialize(struct plugin *pl, char *host, char *port)
{
	struct plugin_dns_data *data = (struct plugin_dns_data*) pl->data;
	int i;

	for (i = 0; i < MAXCLIENTS; i++)
		conn_init(&data->conns[i]);
	data->conn = data->conns;
	memset(&data->q, 0, sizeof(data->q));
	if (server) {
		data->fd = udp_open(port);
	} else {
		data->fd = udp_connect(host, port);
	}

	if (data->fd == -1) 
		return -1;

	event_set(&data->ev, data->fd, EV_PERSIST | EV_READ,
		  plugin_receive, pl);
	event_add(&data->ev, NULL);

	if (server) {
		data->state = PLUGIN_STATE_INITIALIZED;
	} else { /* client */
		data->state = PLUGIN_STATE_CONNECTED;
		
		/* the client should send keep-alive request to the server */
		evtimer_set(&data->timer_ev, plugin_dns_timer_ev_handler, pl);
		register_timer_ev(&data->timer_ev);
		
		data->q.id = 0;
	}

    return 0;
}

void
plugin_deinitialize(struct plugin *pl)
{
	struct plugin_dns_data *data = pl->data;
	
	if (!server) /* client */
		evtimer_del(&data->timer_ev);
	
	event_del(&data->ev);
	close(data->fd);
	data->state = PLUGIN_STATE_UNINITIALIZED;
}

void
plugin_receive(int fd, short ev_type, void *arg)
{
	struct plugin *pl = (struct plugin *) arg;
	struct plugin_dns_data *data = pl->data;
	int n = 0;
	char  buf[PACKETLEN];
	int buflen = PACKETLEN;
	char  packet[PACKETLEN];
	uint8_t clid;
	int conn_flag;
	struct sockaddr_storage from;
	socklen_t fromlen = sizeof(from);
	int r;
	char *domain;
	
	memset(&data->q, 0, sizeof(data->q));

	/* read data from the socket */
	fromlen = sizeof(from);
	r = recvfrom(data->fd, packet, sizeof(packet), 0,
		     (struct sockaddr *) &from, &fromlen);
	if (r == -1) {
		warn("plugin_dns: plugin_receive: recvfrom() returned %d", r);
		plugin_report(pl, 0, REPORT_ERROR_RECEIVE);
		return;
	}

	if (server) {		
		memcpy(&data->addr, &from, fromlen);
		data->addrlen = fromlen;

		/* decode data from the DNS query */
		dns_decode(buf, buflen, &data->q, QR_QUERY, packet, r);
		domain = strstr(data->q.name, data->topdomain);
		n = decode_data(buf, buflen, data->q.name, domain);

		if (n <= 0)
			return;

		/* update the client IP address and UDP port */
		data->addrlen = fromlen;
		memcpy((struct sockaddr*)&data->addr,
		    (struct sockaddr*)&from, fromlen);

		/* use connection 0 as a temporary connection */
		data->conn = &data->conns[0];
		plugin_conn_close(pl, data->conn->clid);
		conn_init(data->conn);
		
		/* payload */
		data->conn->data_sent_after_last_receive = 0;
		process_data_from_plugin(pl, buf, n, &clid, &conn_flag);
		
		/* conn_map() called here */
		
		/* use the reply to send data */
		plugin_send_queued_data(pl, data->conn);
				
	} else { /* client */
		/* check whether the repsonse is coming from the server */
/* 		if (memcmp(&data->addr, &from, data->addrlen) != 0) { */
/* 			printf("Discarding traffic from a different " */
/* 			    "IP address\n"); */
/* 			return; */
/* 		} */

		/* process the response */
		buflen = dns_decode(buf, sizeof(buf), &data->q, QR_ANSWER,
			     packet, r);
		process_data_from_plugin(pl, buf, buflen, &clid, &conn_flag);

		/* 
		 * The client should send back an empty request if there was no
		 * data to send. This allows the server to send more data back.
		 */
		if (!data->conn->data_sent_after_last_receive) {
			plugin_send_empty_request(pl, data->conn);
		}

		return;
	}
}

static void
plugin_send_empty_request(struct plugin *pl, struct conn *conn)
{
	plugin_custom_send(pl, conn->clid, NULL, 0);
}

/* server only */
static void
plugin_send_queued_data(struct plugin *pl, struct conn *conn)
{
/* 	printf("plugin_dns: plugin_send_queued_data(): clid: %hhu,\n" */
/* 	       "\turgent_data: %d, normal_data: %d\n", */
/* 	       conn->clid, */
/* 	       conn->queued_urgent_data_len, conn->queued_normal_data_len); */

	/* if no data was queued then ask the daemon for more data */
	if (conn->queued_urgent_data == NULL &&
	    conn->queued_normal_data == NULL)
		plugin_report(pl, conn->clid, REPORT_READY_TO_SEND);
	
/* 	printf("after plugin_report(READY_TO_SEND)...\n" */
/* 	       "\turgent_data: %d, normal_data: %d\n", */
/* 	       conn->queued_urgent_data_len, conn->queued_normal_data_len); */

	/* send the queued data */
	if (conn->queued_urgent_data != NULL ||
	    conn->queued_normal_data != NULL) {
		if (conn->queued_urgent_data != NULL) { /* urgent data */
			dns_send(pl->data, conn,
			    conn->queued_urgent_data,
			    conn->queued_urgent_data_len);
			free(conn->queued_urgent_data);
			conn->queued_urgent_data = NULL;
			conn->queued_urgent_data_len = 0;
		} else { /* normal data */
			dns_send(pl->data, conn,
			    conn->queued_normal_data,
			    conn->queued_normal_data_len);
			free(conn->queued_normal_data);
			conn->queued_normal_data = NULL;
			conn->queued_normal_data_len = 0;
		}
	}
/* 	printf("after send_dns_pkt()...\n" */
/* 	       "\turgent_data: %d, normal_data: %d\n", */
/* 	       conn->queued_urgent_data_len, conn->queued_normal_data_len); */
}

int
plugin_is_ready_to_send(struct plugin *pl, clientid_t clid)
{
	struct plugin_dns_data *data = pl->data;
	struct conn *conn = conn_by_clid(data, clid);
	
	if (server) {
		if (conn->queued_normal_data == NULL)
			return WILL_QUEUE;
		else
			return QUEUE_FULL;
	} else /* client */
		return WILL_SEND_IMMEDIATELY;
}

void
plugin_conn_close(struct plugin *pl, uint8_t clid)
{
	struct plugin_dns_data *data = pl->data;
	struct conn *conn = conn_by_clid(data, clid);
	
	conn->used = 0;
	if (conn->queued_normal_data != NULL) {
		free(conn->queued_normal_data);
		conn->queued_normal_data = NULL;
	}

	if (conn->queued_urgent_data != NULL) {
		free(conn->queued_urgent_data);
		conn->queued_urgent_data = NULL;
	}
}

/* server only */
void
plugin_conn_map(struct plugin *pl, uint8_t clid, int conn_flag)
{
	struct plugin_dns_data *data = pl->data;
	struct conn *conn = conn_by_clid(data, clid);

	if (!server) /* client */
		return;

	if (conn_flag == CONN_PERM && clid != 0) {
		if (conn->used == 0) {
			printf("plugin_send: CONN_PERM, clid: %hhd, "
			    "updating...\n", clid);
			conn_init(conn);
			conn->used = 1;
		}
		conn->clid = clid;
		data->conn = conn;
	} else if (conn_flag == CONN_TEMP && clid != 0) {
		data->conn = conn;
	}
}

int
plugin_send(struct plugin *pl, clientid_t clid,
   char *data, int len, int data_type, int *consumed)
{
	struct plugin_dns_data *pldata = pl->data;
	char **queued_data;
	int *queued_data_len;
	struct conn *conn;

	conn = conn_by_clid(pldata, clid);

	if (server) { /* server  - queue the data */
		/* locate the queue corresponding to the data type */
		switch (data_type) {
		case NORMAL_DATA:
			queued_data = &conn->queued_normal_data;
			queued_data_len = &conn->queued_normal_data_len;
			break;
			
		case URGENT_DATA:
			queued_data = &conn->queued_urgent_data;
			queued_data_len = &conn->queued_urgent_data_len;
			break;
			
		default:
			errx(EX_SOFTWARE, "plugin_dns: plugin_send(): "
			     "invalid data_type 0x%x\n", data_type);
			return (SEND_ERROR);
		}
		
		/* queue the data */
		if (*queued_data == NULL) {
			*queued_data = malloc(len);
			if (*queued_data == NULL) {
				return (SEND_ERROR_MALLOC);
			}
			memcpy(*queued_data, data, len);
			*queued_data_len = len;
			*consumed = len;
			return (SEND_PKT_QUEUED);
		} else {
			*consumed = 0;
			return (SEND_ERROR_QUEUE_FULL);
		}

	} else { /* client - send the data straight away */
		pldata->q.id++;
		*consumed = dns_send(pldata, pldata->conns, data, len);
		if (*consumed > 0) {
			conn->data_sent_after_last_receive = 1;
			return (SEND_PKT_SENT);
		} else
			return (SEND_ERROR);
	}
}

void
plugin_custom_receive(struct plugin *pl, uint8_t clid,
		      char *data, int len)
{
	plugin_send_queued_data(pl, conn_by_clid(pl->data, clid));
}
