/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ICMP encapsulation plugin
 *
 * Client encapsulates the data as ICMP payload of ICMP echo
 * requests. The server encapsulates the data as ICMP payload of
 * replies to the received ICMP echo requests. For this reasons, the
 * server always queues one packet at a time from the daemon.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/time.h>
#include <err.h>
#include <errno.h>
#include <event.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include "mtund.h"
#include "plugin.h"

/*
 * how often should an empty request be sent to the server - This is
 * useful when the server has data to send but the client doesn't.
 */
#define PLUGIN_ICMP_KEEP_ALIVE_SEC 0
#define PLUGIN_ICMP_KEEP_ALIVE_USEC 500000

struct conn {
	clientid_t clid;
	struct sockaddr_storage addr;
	socklen_t	 addrlen;
	u_int16_t	 id;  /* ICMP echo ID */
	u_int16_t	 seq; /* ICMP echo SEQ */
	int		 used;

	/* normal data queue - regular traffic from the tun interface
	 * and ping requests */
	char		*queued_normal_data;
	int		 queued_normal_data_len;
	/* urgent data queue - ping replies */
	char		*queued_urgent_data;
	int		 queued_urgent_data_len;
	/*
	 * Has the client sent data after the last reply from the
	 * server was received?
	 */
	int		 data_sent_after_last_receive; /* client only */
};

struct plugin_icmp_data {
	int		 fd; /* icmp socket */
	int		 state;

	/* previous value of net.inet.icmp.echo_user sysctl variable */
	int		 old_sysctl;
	size_t		 old_sysctl_size;

	struct event	 ev; /* used by libevent to monitor our fd */
	struct event	 timer_ev;
	struct conn	 conns[MAXCLIENTS];
	struct conn	*conn; /* most recently receiving conn */
};

/* ICMP echo header */
struct my_icmp_hdr {
        u_int8_t        type;
        u_int8_t        code;
        u_int16_t       cksum;
        u_int16_t       id;
        u_int16_t       seq;
};


static void plugin_send_queued_data();
static void plugin_send_empty_request(struct plugin *pl, struct conn *conn);

static u_int16_t
in_cksum(u_int16_t *addr, int len)
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	while (nleft > 1) {
		sum += *w++;
		nleft -= 2;
	}

	if (nleft == 1)
		*(u_char *) (&answer) = *(u_char *) w; sum += answer;

	sum = (sum >> 16) + (sum & 0xffff);   /* add hi 16 to low 16 */
	sum += (sum >> 16);           /* add carry */
	answer = ~sum;                /* truncate to 16 bits */

	return (answer);
}

static void
conn_init(struct conn *conn)
{
	conn->used = 0;

	conn->id = 0;
	conn->seq = 0;

	conn->queued_normal_data = NULL;
	conn->queued_normal_data_len = 0;
	conn->queued_urgent_data = NULL;
	conn->queued_urgent_data_len = 0;	
}

static struct conn*
conn_by_clid(struct plugin_icmp_data *data, clientid_t clid)
{
	if (server)
		if (clid == 0)
			return data->conn;
		else
			return &data->conns[clid];
	else
		return data->conns;
}

int
plugin_register(struct plugin* pl)
{
    pl->name = "icmp";
    pl->mtu = 1400;
    pl->initialize = plugin_initialize;
    pl->deinitialize = plugin_deinitialize;
    pl->is_ready_to_send = plugin_is_ready_to_send;
    pl->send = plugin_send;
    pl->receive = plugin_receive;
    pl->conn_close = plugin_conn_close;
    pl->conn_map = plugin_conn_map;
    pl->custom_receive = plugin_custom_receive;

    pl->data = malloc(sizeof(struct plugin_icmp_data));
    if (!pl->data) {
	return -1;
    }
    struct plugin_icmp_data *data = (struct plugin_icmp_data*) pl->data;
    data->fd = -1;
    data->state = PLUGIN_STATE_UNINITIALIZED;

    return 0;
}

void
plugin_unregister(struct plugin* pl)
{
    if (pl->data) {
	free(pl->data);
	pl->data = NULL;
    }
}

/* register a timer event */
static void
register_timer_ev(struct event *ev)
{
    struct timeval tv;
    
    tv.tv_sec=PLUGIN_ICMP_KEEP_ALIVE_SEC;
    tv.tv_usec=PLUGIN_ICMP_KEEP_ALIVE_USEC;
    evtimer_del(ev);
    evtimer_add(ev, &tv);
}

/* handler function for the libevent timer event - client only */
void 
plugin_icmp_timer_ev_handler(int fd, short ev_type, void *arg)
{
    struct plugin_icmp_data *data = ((struct plugin *)arg)->data;

    /* send a request to the server */
    if (server == 0) { /* client */
	    plugin_send_empty_request(arg, data->conn);
	    //plugin_receive(data->fd, EV_TIMEOUT, arg);
    }

     /* register a timer event again */
    register_timer_ev(&data->timer_ev);
}

int
plugin_initialize(struct plugin *pl, char *host, char *port)
{
	struct addrinfo hints, *ai;
	int i, n;
	int new_sysctl = 1;
	int fd_flags;
	
	struct plugin_icmp_data *data = (struct plugin_icmp_data*) pl->data;
	
	conn_init(data->conns);
	data->conn = data->conns;
	if (server) {
		for (i = 0; i < MAXCLIENTS; i++)
			conn_init(&data->conns[i]);
		
		data->conns->id = 0;
		data->conns->seq = 0;
	} else {
		srandomdev();
		data->conns->id = random();
		data->conns->seq = random();

		memset(&hints, 0, sizeof(hints));
		hints.ai_family = AF_INET;
		hints.ai_socktype = SOCK_RAW;
		
		n = getaddrinfo(host, NULL, &hints, &ai);
		if (n) {
			fprintf(stderr, "getaddrinfo failed: %s\n",
			    gai_strerror(n));
			return -1;
		}
		if (ai) {
			memcpy(&data->conns->addr, ai->ai_addr,
			    ai->ai_addrlen);
			data->conns->addrlen = ai->ai_addrlen;
		} else {
			fprintf(stderr, "bind failed for port %s\n", port);
			return -1;
		}
		freeaddrinfo(ai);
	}

	/* open the socket */
	data->fd = socket (AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (data->fd == -1) 
		return -1;

	/* non-blocking i/o */
	fd_flags = fcntl(data->fd, F_GETFL, 0);
	if (fd_flags == -1)
		errx(EX_OSFILE,
		    "Failed to get flags from the icmp socket fd\n");
	fcntl(data->fd, F_SETFL, fd_flags|O_NONBLOCK);

	event_set(&data->ev, data->fd, EV_PERSIST | EV_READ,
		  plugin_receive, pl);
	event_add(&data->ev, NULL);

	if (server) {
#ifdef __FreeBSD__
		data->old_sysctl_size = sizeof(data->old_sysctl);
		if (0 != sysctlbyname("net.inet.icmp.echo_user",
			     &data->old_sysctl, &data->old_sysctl_size,
			     &new_sysctl, sizeof(new_sysctl)))
			errx(EX_UNAVAILABLE, "Cannot set net.inet.icmp."
			    "echo_user sysctl. Maybe you need to patch "
			    "your kernel?");
#endif
#ifdef __linux__
		system("sysctl net.ipv4.icmp_echo_ignore_all=0");
#endif

		data->state = PLUGIN_STATE_INITIALIZED;

	} else { /* client */
		data->state = PLUGIN_STATE_CONNECTED;

		/* the client should send keep-alive request to the server */
		evtimer_set(&data->timer_ev, plugin_icmp_timer_ev_handler, pl);
		register_timer_ev(&data->timer_ev);
    }

    return 0;
}

void
plugin_deinitialize(struct plugin *pl)
{
	struct plugin_icmp_data *data = pl->data;
	
	event_del(&data->ev);
	close(data->fd);
	data->state = PLUGIN_STATE_UNINITIALIZED;
	
#ifdef __FreeBSD__
	if (server)
		sysctlbyname("net.inet.icmp.echo_user", NULL, 0,
		    &data->old_sysctl, data->old_sysctl_size);
#endif
}

static int
send_icmp_pkt(struct plugin *pl, struct conn *conn, char *data, int len)
{
	struct plugin_icmp_data *datapl = pl->data;
	int n = 0;
	char* packet;
	struct my_icmp_hdr *icmp;
	
	printf("send_icmp_pkt(): len: %d\n", len);
	
	packet = malloc(sizeof(struct my_icmp_hdr) + len);
	if (!packet) {
		fprintf(stderr, "plugin_send: out of memory\n");
		return SEND_ERROR_MALLOC;
	}
	memset (packet, 0, sizeof(struct my_icmp_hdr) + len);
	memcpy (packet+sizeof(struct my_icmp_hdr), data, len);
	
	icmp = (struct my_icmp_hdr*)(packet);
	if (server) {
		icmp->type = ICMP_ECHOREPLY;
	} else { /* client */
		icmp->type = ICMP_ECHO;
		//icmp->type = 123;	
	}
	
	icmp->code = 0;
	icmp->id = htons(conn->id);
	icmp->seq = htons(conn->seq);
	icmp->cksum = 0;
	icmp->cksum = in_cksum((u_int16_t*)packet,
			  sizeof(struct my_icmp_hdr) + len);
	
	n = sendto(datapl->fd, (char*)packet,
		sizeof(struct my_icmp_hdr) + len, 0,
		(struct sockaddr*)&conn->addr, conn->addrlen);
	//(struct sockaddr*)dst, sizeof (struct sockaddr_in));
	fprintf(stderr, "plugin_send: send returned %d\n", n);
	
	/* the client has to reset the timer for keep-alive requests */
	if (!server) { /* client */
		/* increment the sequence number */
		if (conn->seq == 65535)
			conn->seq = 0;
		else 
			conn->seq++;

		register_timer_ev(&datapl->timer_ev);
	}
	
	return n;
}

void
plugin_receive(int fd, short ev_type, void *arg)
{
	struct plugin *pl = (struct plugin *) arg;
	struct plugin_icmp_data *data = pl->data;
	int n = 0;
	char  packet[PACKETLEN];
	char *packetp = packet;
	struct sockaddr_in from;
	socklen_t fromlen = sizeof(from);
	struct my_icmp_hdr *icmp = NULL;
	uint8_t clid;
	int conn_flag;
	
	printf("plugin_receive(ev_type: 0x%x)\n", ev_type);
	
	/* read data from the socket */
	n = recvfrom(data->fd, packet, sizeof(packet), 0,
		     (struct sockaddr *) &from, &fromlen);
	if (n == -1) {
		if (errno == EAGAIN)
			goto send_request;

		warn("plugin_cimp: plugin_receive(): recvfrom() returned %d",
		     n);
		plugin_report(pl, 0, REPORT_ERROR_RECEIVE);
		return;
	}
	
	/* use connection 0 as a temporary connection */
	data->conn = data->conns;
	if (server) {
		conn_init(data->conn);
		memcpy(&data->conn->addr, &from, fromlen);
		data->conn->addrlen = fromlen;
	} else { /* client */
		/* the server cannot change the address */
		if (memcmp(&data->conn->addr, &from, fromlen) != 0)
			goto pkt_not_for_us;
	}
	
	/* process the data */
	/* check whether the packet is for us */
	if (n < 20 + sizeof(struct my_icmp_hdr)) {
		warnx("plugin_icmp: plugin_receive: received data too short "
		      "to contain IP and ICMP headers\n");
		goto pkt_not_for_us;
	}
	
	/* IP header */
	packetp += 20;
	n -= 20;
	
	/* ICMP header */
	icmp = (void*)packetp;
	packetp += sizeof(struct my_icmp_hdr);
	n -= sizeof(struct my_icmp_hdr);
	data->conn->id = ntohs(icmp->id);
	data->conn->seq = ntohs(icmp->seq);

	/* check values in the ICMP header */
	if (server) {
		if (icmp->type != ICMP_ECHO) // or 123
			goto pkt_not_for_us;
	} else {
		if (icmp->type != ICMP_ECHOREPLY)
			goto pkt_not_for_us;
	}
	
	/* payload */
	data->conn->data_sent_after_last_receive = 0;
	if (n > 0)
		process_data_from_plugin(pl, packetp, n, &clid, &conn_flag);
	
	/* conn_map() called here */
	
	/* use the reply to send data */
	if (server)
		plugin_send_queued_data(pl, data->conn);

	/* 
	 * The client should send back an empty request if there was no
	 * data to send. This allows the server to send more data back.
	 */
 send_request:
	if (!server && !data->conn->data_sent_after_last_receive) {
		plugin_send_empty_request(pl, data->conn);
	}
	return;
	
 pkt_not_for_us:
	fprintf(stderr, "discarding data from a different client\n");
}

static void
plugin_send_empty_request(struct plugin *pl, struct conn *conn)
{
	plugin_custom_send(pl, conn->clid, NULL, 0);
}

/* server only */
static void
plugin_send_queued_data(struct plugin *pl, struct conn *conn)
{
	printf("plugin_icmp: plugin_send_queued_data(): clid: %hhu,\n"
	       "\turgent_data: %d, normal_data: %d\n",
	       conn->clid,
	       conn->queued_urgent_data_len, conn->queued_normal_data_len);

	/* if no data was queued then ask the daemon for more data */
	if (conn->queued_urgent_data == NULL &&
	    conn->queued_normal_data == NULL)
		plugin_report(pl, conn->clid, REPORT_READY_TO_SEND);
	
	printf("after plugin_report(READY_TO_SEND)...\n"
	       "\turgent_data: %d, normal_data: %d\n",
	       conn->queued_urgent_data_len, conn->queued_normal_data_len);

	/* send the queued data */
	if (conn->queued_urgent_data != NULL ||
	    conn->queued_normal_data != NULL) {
		if (conn->queued_urgent_data != NULL) {
			send_icmp_pkt(pl, conn,
			    conn->queued_urgent_data,
			    conn->queued_urgent_data_len);
			free(conn->queued_urgent_data);
			conn->queued_urgent_data = NULL;
			conn->queued_urgent_data_len = 0;
		} else {
			send_icmp_pkt(pl, conn,
			    conn->queued_normal_data,
			    conn->queued_normal_data_len);
			free(conn->queued_normal_data);
			conn->queued_normal_data = NULL;
			conn->queued_normal_data_len = 0;
		}
	}
	printf("after send_icmp_pkt()...\n"
	       "\turgent_data: %d, normal_data: %d\n",
	       conn->queued_urgent_data_len, conn->queued_normal_data_len);
}

int
plugin_is_ready_to_send(struct plugin *pl, clientid_t clid)
{
	struct plugin_icmp_data *data = pl->data;
	struct conn *conn = conn_by_clid(data, clid);
	
	if (server) {
		if (conn->queued_normal_data == NULL)
			return WILL_QUEUE;
		else
			return QUEUE_FULL;
	} else /* client */
		return WILL_SEND_IMMEDIATELY;
}

void
plugin_conn_close(struct plugin *pl, uint8_t clid)
{
	struct plugin_icmp_data *data = pl->data;
	struct conn *conn = conn_by_clid(data, clid);
	
	conn->used = 0;
	if (conn->queued_normal_data != NULL) {
		free(conn->queued_normal_data);
		conn->queued_normal_data = NULL;
	}

	if (conn->queued_urgent_data != NULL) {
		free(conn->queued_urgent_data);
		conn->queued_urgent_data = NULL;
	}
}

/* server only */
void
plugin_conn_map(struct plugin *pl, uint8_t clid, int conn_flag)
{
	struct plugin_icmp_data *data = pl->data;
	struct conn *conn = conn_by_clid(data, clid);

	if (!server) /* client */
		return;

	if (conn_flag == CONN_PERM && clid != 0) {
		if (conn->used == 0) {
			printf("plugin_send: CONN_PERM, clid: %hhd, "
			    "updating...\n", clid);
			conn_init(conn);
			conn->used = 1;
		}
		conn->clid = clid;
		//memcpy(&conn->addr, &data->conn->addr, data->conn->addrlen);
		memcpy(&conn->addr, &data->conn->addr, sizeof(conn->addr));
		conn->addrlen = data->conn->addrlen;
		conn->id = data->conn->id;
		conn->seq = data->conn->seq;
		data->conn = conn;
	}
	if (conn_flag == CONN_TEMP && clid != 0) {
		conn->id = data->conn->id;
		conn->seq = data->conn->seq;
		data->conn = conn;
	}
}

int
plugin_send(struct plugin *pl, clientid_t clid,
    char *data, int len, int data_type, int *consumed)
{
	struct plugin_icmp_data *pldata = pl->data;
	int n = 0;
	char **queued_data;
	int *queued_data_len;
	struct conn *conn;

	conn = conn_by_clid(pldata, clid);

	printf("plugin_icmp: plugin_send(): clid: %hhu, data_type: %d\n",
	       clid, data_type);
	printf("plugin_send(): urgent_data: %d, normal_data: %d\n",
	       conn->queued_urgent_data_len, conn->queued_normal_data_len);
	
/* 	if (pldata->state != PLUGIN_STATE_CONNECTED) { */
/* 		fprintf(stderr, "no client connected yet, discarding data\n"); */
/* 		return SEND_ERROR; */
/* 	} */
	
	if (server) { /* server  - queue the data */
		switch (data_type) {
		case NORMAL_DATA:
			queued_data = &conn->queued_normal_data;
			queued_data_len = &conn->queued_normal_data_len;
			goto process_queued_data;
			
		case URGENT_DATA:
			queued_data = &conn->queued_urgent_data;
			queued_data_len = &conn->queued_urgent_data_len;

		process_queued_data:
			if (*queued_data == NULL) {
				*queued_data = malloc(len);
				if (*queued_data == NULL) {
					return SEND_ERROR_MALLOC;
				}
				memcpy(*queued_data, data, len);
				*queued_data_len = len;
				*consumed = len;
				return SEND_PKT_QUEUED;
			} else {
				*consumed = 0;
				return SEND_ERROR_QUEUE_FULL;
			}
			
		default:
			errx(EX_SOFTWARE, "plugin_icmp: plugin_send(): "
			     "invalid data_type 0x%x\n", data_type);
		}

	} else { /* client - send the data straight away */
		n = send_icmp_pkt(pl, pldata->conns, data, len);
		fprintf(stderr, "send_icmp_pkt: send returned %d\n", n);
		if (n > 0) {
			*consumed = n;
			conn->data_sent_after_last_receive = 1;
			return SEND_PKT_SENT;
		} else
			*consumed = 0;
			return SEND_ERROR;
	}
}

void
plugin_custom_receive(struct plugin *pl, uint8_t clid,
		      char *data, int len)
{
	plugin_send_queued_data(pl, conn_by_clid(pl->data, clid));
}
