/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <arpa/inet.h>
#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>
#include <event.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include "mtund.h"
#include "plugin.h"

#define MAXTMPCONNS 10
#define TEMP_CONN_TIMEOUT 5
#define USE_TCP_LISTENALL 1

#define min(a,b) ( (a>b) ? b : a )

/* structure representing a connection, one per client */
struct conn {
	int status;
	uint8_t clid;
	int fd;
	struct event ev;
	struct event timer_ev;
	struct plugin_tcp_data *data;
};

enum {
	CONN_STATUS_FREE,	/* connection not in use */
	CONN_STATUS_TEMP,	/* temporary connection */
	CONN_STATUS_PERM	/* permanenet connection */
};

/* plugin-specific data */
struct plugin_tcp_data {
	int state; /* is a client connected? */
	/* conns[0]: server - bind fd, client - connect fd */
	struct conn conns[MAXCLIENTS]; /* server - accept fds */
	struct conn tmpconns[MAXTMPCONNS]; /* server - accept fds */
	struct conn *conn; /* most recently receiving connection */
	struct plugin *pl;
};

void plugin_accept_new_conn(int fd, short ev_type, void *arg);
void plugin_conn_timeout(int fd, short ev_type, void *arg);
static void conn_discard(struct conn *conn);
void plugin_conn_close(struct plugin *pl, uint8_t clid);
static void conn_update_to_perm(struct plugin_tcp_data *data,
		uint8_t clid, int fd);

/*
 * Create a listening TCP endpoint. First get the list of potential
 * network layter addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create a
 * listening TCP endpoint is successful (or no other alternative
 * exists).
 */

static int
tcp_listen(char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, on = 1;
    
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(NULL, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "getaddrinfo failed: %s\n",
                gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }

        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
        if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        fprintf(stderr, "bind failed for port %s\n",
                port);
        return -1;
    }

    if (listen(fd, 1) < 0) {
        fprintf(stderr, "listen failed: %s\n", strerror(errno));
	close(fd);
        return -1;
    }

    return fd;
}

/*
 * Accept a new TCP connection and write a message about who was
 * accepted to the system log.
 */

static int
tcp_accept(int listen)
{
    struct sockaddr_storage ss;
    socklen_t ss_len = sizeof(ss);
    char host[NI_MAXHOST];
    char serv[NI_MAXSERV];
    int n, fd;
    fd = accept(listen, (struct sockaddr *) &ss, &ss_len);
    if (fd == -1) {
        syslog(LOG_ERR, "accept failed: %s", strerror(errno));
        return -1;
    }

    n = getnameinfo((struct sockaddr *) &ss, ss_len,
                    host, sizeof(host), serv, sizeof(serv),
                    NI_NUMERICHOST);
    if (n) {
        fprintf(stderr, "getnameinfo failed: %s", gai_strerror(n));
    } else {
        fprintf(stderr, "connection from %s:%s", host, serv);
    }

    return fd;
}

/*
 * Establish a connection to a remote TCP server. First get the list
 * of potential network layer addresses and transport layer port
 * numbers. Iterate through the returned address list until an attempt
 * to establish a TCP connection is successful (or no other
 * alternative exists).
 */

static int
tcp_connect(const char *host, const char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, serrs = 0, cerrs = 0;
    struct timeval tv;
    
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(host, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            serrs++;
            continue;
        }

	tv.tv_sec  = SOCKET_TIMEOUT;
	tv.tv_usec = 0;
	setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv));

        if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        cerrs++;
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        if ((serrs == 1 && ! cerrs) || (!serrs && cerrs == 1)) {
            fprintf(stderr, "%s: %s\n",
                    serrs ? "socket" : "connect", strerror(errno));
        } else {
            fprintf(stderr, "socket or connect: failed for %s port %s\n",
                    host, port);
        }
        return -1;
    }

    return fd;
}
/*
 * Close a TCP connection. This function trivially calls close() on
 * POSIX systems, but might be more complicated on other systems.
 */

static int
tcp_close(int fd)
{
    return close(fd);
}

/* timer event handler function - used to timeout CONN_TYPE_TEMP connections */
void
plugin_conn_timeout(int fd, short ev_type, void *arg)
{
	conn_discard(arg);
}

/* close connection */
static void
conn_discard(struct conn *conn)
{
	if (conn->fd != -1) {
		tcp_close(conn->fd);
		event_del(&conn->ev);
		if (conn->status == CONN_STATUS_TEMP)
			evtimer_del(&conn->timer_ev);
	}
	conn->fd = -1;
	conn->status = CONN_STATUS_FREE;
}

/* close connection - calls conn_discard() */
void
plugin_conn_close(struct plugin *pl, uint8_t clid)
{
	struct plugin_tcp_data *data = pl->data;
	if (server)
		if (clid == 0)
			conn_discard(data->conn);
		else
			conn_discard(&data->conns[clid]);
	else
		conn_discard(data->conns);
}

/* server only - update connection to CONN_TYPE_PERM */
static void
conn_update_to_perm(struct plugin_tcp_data *data, uint8_t clid, int fd)
{
	struct conn *conn = &data->conns[clid];
	
	if (conn->clid == clid && conn->fd == fd)
		return; /* nothing to do */
	
	/* discard previous connection */
	if (conn->status == CONN_STATUS_PERM)
		conn_discard(conn);
	
	/* set up the new connection */
	conn->fd = fd;
	conn->clid = clid;
	event_set(&conn->ev, conn->fd, EV_PERSIST | EV_READ,
	    plugin_receive, conn);
	event_add(&conn->ev, NULL);
	evtimer_set(&conn->timer_ev, plugin_conn_timeout, conn);
}

/* server only */
void
plugin_conn_map(struct plugin *pl, uint8_t clid, int conn_flag)
{
	struct plugin_tcp_data *data = pl->data;
	struct conn *conn = data->conn;
	int fd;

	if (!server)
		return;

	/* update the connection status (server only) */
	switch (conn_flag) {
	case CONN_DISCARD:
		conn_discard(conn);
			return;
			
	case CONN_TEMP:
		break;
		
	case CONN_PERM:
		/* migrate a temporary connection to a permanent one */
		fd = conn->fd;
		if (conn->status == CONN_STATUS_TEMP) {
			conn->status = CONN_STATUS_FREE;
			evtimer_del(&conn->timer_ev);
			event_del(&conn->ev);
			conn->fd = -1;
		}
		conn_update_to_perm(data, clid, fd);
		break;
	}
}	


int
plugin_register(struct plugin* pl)
{
    pl->name = "tcp";
    pl->mtu = 1450;
    pl->initialize = plugin_initialize;
    pl->deinitialize = plugin_deinitialize;
    pl->is_ready_to_send = plugin_is_ready_to_send;
    pl->send = plugin_send;
    pl->receive = plugin_receive;
    pl->conn_close = plugin_conn_close;
    pl->conn_map = plugin_conn_map;
    pl->custom_receive = plugin_custom_receive;

    pl->data = malloc(sizeof(struct plugin_tcp_data));
    if (!pl->data)
	    return -1;
    struct plugin_tcp_data *data = pl->data;
    data->pl = pl;
    data->state = PLUGIN_STATE_UNINITIALIZED;
    return 0;
}

void
plugin_unregister(struct plugin* pl)
{
    if (pl->data) {
	free(pl->data);
	pl->data = NULL;
    }
}

int
plugin_initialize(struct plugin *pl, char *host, char *port)
{
    struct plugin_tcp_data *data = (struct plugin_tcp_data*) pl->data;
    int i;

    data->pl = pl;

    /* initialize conns and tmpconns */
    for (i = 0; i < MAXCLIENTS; i++) {
	    data->conns[i].status = CONN_STATUS_FREE;
	    data->conns[i].fd = -1;
	    data->conns[i].data = data;
    }
    for (i = 0; i < MAXTMPCONNS; i++) {
	    data->tmpconns[i].status = CONN_STATUS_FREE;
	    data->tmpconns[i].fd = -1;
	    data->tmpconns[i].data = data;
	    data->tmpconns[i].clid = 0;
    }
    data->conn = NULL;
    
    /* open the socket */
    if (server) {
	    data->conns->fd = tcp_listen(port);
	    /* enable TCP_LISTENALL */
	    if (USE_TCP_LISTENALL != 0) {
		    int soval = 1;
		    if (0 != setsockopt(data->conns->fd, IPPROTO_TCP,
				 TCP_LISTENALL, &soval, sizeof(soval)))
			    warnx("setsockopt(TCP_LISTENALL) failed\n");
	    }

	    event_set(&data->conns->ev, data->conns->fd, EV_PERSIST | EV_READ,
		plugin_accept_new_conn, data->conns);
	    if (data->conns->fd != -1) {
		    data->state = PLUGIN_STATE_INITIALIZED;
		    event_add(&data->conns->ev, NULL);
	    }
    } else { /* client */
	    data->conns->fd = tcp_connect(host, port);
	    event_set(&data->conns->ev, data->conns->fd, EV_PERSIST | EV_READ,
		plugin_receive, data->conns);
	    if (data->conns->fd != -1) {
		    data->state = PLUGIN_STATE_CONNECTED;
		    event_add(&data->conns->ev, NULL);
	    }
    }
    data->conns->status = CONN_STATUS_PERM;

    if (data->conns->fd != -1) {
	    return 0;
    } else
	    return -1;
}

void
plugin_deinitialize(struct plugin *pl)
{
	int i;
	struct plugin_tcp_data *data = (struct plugin_tcp_data*) pl->data;
	
	/* deinitialize all connections */
	for (i = 0; i < MAXCLIENTS; i++)
		conn_discard(&data->conns[i]);
	for (i = 0; i < MAXTMPCONNS; i++)
		conn_discard(&data->tmpconns[i]);
	data->conn = NULL;

	data->state = PLUGIN_STATE_UNINITIALIZED;
}

/* accept a new connection on the listening fd - called by libevent */
void
plugin_accept_new_conn(int fd, short ev_type, void *arg)
{
	struct plugin_tcp_data *data = ((struct conn *)arg)->data;
	struct conn *conn = NULL;
	int i;
	struct timeval tv;

/* 	assert(server); */
/* 	assert(conn == data->conn); */

	/* find a free tmpcopnn to store the connection metadata */
	for (i = 0; i < MAXTMPCONNS; i++)
		if (data->tmpconns[i].status == CONN_STATUS_FREE) {
			conn = &data->tmpconns[i];
			break;
		}
	
	if (conn == NULL) {
		printf("plugin_tcp: too many new connections\n");
		return;
	}

	conn->fd = tcp_accept(fd);
	if (conn->fd == -1) {
		printf("accept failed\n");
		plugin_report(data->pl, 0, REPORT_ERROR_RECEIVE);
	} else {
		printf("new client connection accepted\n");
		conn->clid = 0;
		conn->status = CONN_STATUS_TEMP;
		event_set(&conn->ev, conn->fd, EV_PERSIST | EV_READ,
			  plugin_receive, conn);
		event_add(&conn->ev, NULL);

		tv.tv_sec = TEMP_CONN_TIMEOUT;
		tv.tv_usec = 0;
		evtimer_set(&conn->timer_ev, plugin_conn_timeout, conn);
		evtimer_add(&conn->timer_ev, &tv);

		data->conn = conn;
		/* data->state = PLUGIN_STATE_CONNECTED; */
	}
}

void
plugin_receive(int fd, short ev_type, void *arg)
{
	struct conn *conn = arg;
	struct plugin_tcp_data *data = conn->data;
	int n = 0;
	char packet[PACKETLEN];
	char *packet_p = packet;
	uint16_t len;
	uint8_t clid = 0;
	int conn_flag;

	debug("plugin_tcp: plugin_receive(): fd: %d, conn: 0x%x, "
	    "conn->fd: %d\n", fd, (uint32_t)conn, conn->fd);

	/* get length of the next packet */
	n = recv(fd, &len, sizeof(len),0);
	if (n <= 0)
		goto recv_error;
	
	/* packet too large */
	if (len >= sizeof(packet)) {
		printf("client sent a too large packet (len: %d)\n", len);
		goto recv_error;
	}

	/* get the packet */
	while (len > 0) {
		n = recv(fd, packet_p, len, 0);
		if (n <= 0)
				goto recv_error;
		len -= n;
		packet_p += n;
	}

	/* update the most recent connection */
	data->conn = arg;
	
	/* pass the packet to the daemon */
	process_data_from_plugin(data->pl, packet, n, &clid, &conn_flag);
	
	return;
 recv_error:
	/* connection broke down or client disconnected */
	conn_discard(conn);
	//plugin_report(data->pl, conn->clid, REPORT_ERROR_RECEIVE);
}

int
plugin_is_ready_to_send(struct plugin *pl, uint8_t clid)
{
	return WILL_SEND_IMMEDIATELY;
}

/*
 * BUG: the write might not send the whole buffer in one go, should check
 * and call again, but then the receiver would have to reassemble packets
 */
int
plugin_send(struct plugin *pl, uint8_t clid,
    char *data, int len, int data_type, int* consumed)
{
	struct plugin_tcp_data *pldata = (struct plugin_tcp_data*) pl->data;
	int n = 0;
	char *ldata;
	uint16_t l = len;
	struct conn *conn;
	
	printf("plugin_tcp: plugin_send(): clid: %hhu\n", clid);

	/* lookup the connection metadata */
	if (server) {
		if (clid == 0)
			conn = pldata->conn;
		else
			conn = &pldata->conns[clid];
	} else /* client */
		conn = pldata->conns;
	
	printf("plugin_tcp: plugin_send(): conn: 0x%x, conn->fd: %d\n",
	       (uint32_t)conn, conn->fd);

	if (conn->fd == -1) {
		printf("connection not initialized yet, discarding data\n");
		return SEND_ERROR;
	}
	
	/* framing - prepend payload length before the payload */
	ldata = malloc(l+sizeof(l));
	if (ldata == NULL)
		return (SEND_ERROR_MALLOC);
	memcpy(ldata, &len, sizeof(l));
	memcpy(ldata + sizeof(l), data, l);
	n = write(conn->fd, ldata, sizeof(l) + l);
	free(ldata);
	
	if (n < 0) {
		*consumed = 0;
		warn("plugin_tcp: write failed");
	} else
		*consumed = n;
	if (n < 0 || (l != 0 && n == 0) ) {
		plugin_report(pl, clid, REPORT_ERROR_SEND);
		return SEND_ERROR;
	}
	
	return SEND_PKT_SENT;
}

void
plugin_custom_receive(struct plugin *pl, uint8_t clid,
		      char *data, int len)
{
	/* nothing to do */
}
