/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/queue.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <arpa/inet.h>
#include <netdb.h>

#include <err.h>
#include <errno.h>
#include <sys/time.h>
#include <event.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "mtund.h"
#include "plugin.h"

/* structure representing a connection, one per client */
struct conn {
	struct sockaddr_storage addr;
	socklen_t		addrlen;
};

/* plugin-specific data */
struct plugin_udp_data {
	int	state;
	int	fd; /* udp socket to the other endpoint */
	struct event ev; /* for monitoring fd */
	struct conn conns[MAXCLIENTS]; /* 0 is last conn and client conn */
};

/*
 * Establish a connected UDP endpoint. First get the list of potential
 * network layer addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create and
 * connect a UDP endpoint is successful (or no other alternative
 * exists).
 */

static int
udp_connect(char *host, char *port)
{
	struct addrinfo hints, *ai_list, *ai;
	int n, fd = 0;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	
	n = getaddrinfo(host, port, &hints, &ai_list);
	if (n) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(n));
		exit(EXIT_FAILURE);
	}
	
	for (ai = ai_list; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd < 0) {
			continue;
		}
		if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
			break;
		}
		close(fd);
	}
	
	freeaddrinfo(ai_list);
	
	if (ai == NULL) {
		fprintf(stderr, "socket or connect: failed for %s port %s\n",
			host, port);
                return -1;
	}
	
	return fd;
}

/*
 * Create a named UDP endpoint. First get the list of potential
 * network layer addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create a UDP
 * endpoint is successful (or no other alternative exists).
 */

static int
udp_open(char *port)
{
	struct addrinfo hints, *ai_list, *ai;
	int n, fd = 0, on = 1;
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	
	n = getaddrinfo(NULL, port, &hints, &ai_list);
	if (n) {
		fprintf(stderr, "getaddrinfo failed: %s\n", gai_strerror(n));
		return -1;
	}
	
	for (ai = ai_list; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd < 0) {
			continue;
		}
		
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
		if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
			break;
		}
		close(fd);
	}
	
	freeaddrinfo(ai_list);
	
	if (ai == NULL) {
		fprintf(stderr, "bind failed for port %s\n", port);
		return -1;
	}
	
	return fd;
}

/*
 * Close a udp socket. This function trivially calls close() on
 * POSIX systems, but might be more complicated on other systems.
 */

static int
udp_close(int fd)
{
    return close(fd);
}


int plugin_register(struct plugin* pl) {
    pl->name = "udp";
    pl->mtu = 1400;
    pl->initialize = plugin_initialize;
    pl->deinitialize = plugin_deinitialize;
    pl->is_ready_to_send = plugin_is_ready_to_send;
    pl->send = plugin_send;
    pl->receive = plugin_receive;
    pl->conn_close = plugin_conn_close;
    pl->conn_map = plugin_conn_map;
    pl->custom_receive = plugin_custom_receive;

    pl->data = malloc(sizeof(struct plugin_udp_data));
    if (!pl->data) {
	return -1;
    }
    struct plugin_udp_data *data = (struct plugin_udp_data*) pl->data;
    data->fd = -1;
    data->state = PLUGIN_STATE_UNINITIALIZED;

    return 0;
}

void plugin_unregister(struct plugin* pl) {
    if (pl->data) {
	free(pl->data);
	pl->data = NULL;
    }
}

int
plugin_initialize(struct plugin *pl, char *host, char *port)
{
	struct plugin_udp_data *data = (struct plugin_udp_data*)pl->data;

	if (server == 1) {
		data->fd = udp_open(port);
		if (data->fd != -1)
			data->state = PLUGIN_STATE_INITIALIZED;
	} else { /* client */
		data->fd = udp_connect(host, port);
		if (data->fd != -1)
			data->state = PLUGIN_STATE_CONNECTED;
	}
	if (data->fd != -1) {
		event_set(&data->ev, data->fd, EV_PERSIST | EV_READ,
			  plugin_receive, pl);
		event_add(&data->ev, NULL);
		return 0;
	} else
		return -1;
}

void
plugin_deinitialize(struct plugin *pl)
{
    struct plugin_udp_data *data = (struct plugin_udp_data*) pl->data;

    event_del(&data->ev);
    udp_close(data->fd);

    data->state = PLUGIN_STATE_UNINITIALIZED;
}

void plugin_receive(int fd, short ev_type, void *arg) {
    struct plugin *pl = (struct plugin *) arg;
    struct plugin_udp_data *data = (struct plugin_udp_data*)pl->data;
    int n = 0;
    char packet[PACKETLEN];
    struct conn *conn = data->conns;
    int conn_flag;
    uint8_t clid = 0;

    conn->addrlen = sizeof(conn->addr);
    n = recvfrom(data->fd, packet, sizeof(packet), 0,
		 (struct sockaddr *) &conn->addr, &conn->addrlen);
    
    if (n == -1) {
	    /* report a broken plugin */
	    plugin_report(pl, 0, REPORT_ERROR_RECEIVE);
	    return;
    }
    
    process_data_from_plugin(pl, packet, n, &clid, &conn_flag);
}

int
plugin_is_ready_to_send(struct plugin *pl, uint8_t clid)
{
	return WILL_SEND_IMMEDIATELY;
}

void
plugin_conn_close(struct plugin *pl, uint8_t clid)
{
	/* nothing to do */
}

/* server only */
void
plugin_conn_map(struct plugin *pl, uint8_t clid, int conn_flag)
{
	struct plugin_udp_data *data = pl->data;

	if (!server)
		return;
	
	if (conn_flag == CONN_PERM && clid != 0)
		memcpy(&data->conns[clid], data->conns, sizeof(struct conn));
}

int
plugin_send(struct plugin *pl, uint8_t clid,
    char *data, int len, int data_type, int* consumed)
{
	struct plugin_udp_data *datapl = (struct plugin_udp_data*) pl->data;
	int nwrite = 0;
	struct conn *conn = &(datapl->conns[clid]);

	if (server) {
		nwrite = sendto(datapl->fd, data, len, 0,
			     (struct sockaddr*)&conn->addr,
			     conn->addrlen);
	} else { /* client */
		nwrite = send(datapl->fd, data, len, 0);
	}
	
	if (nwrite == len) {
		*consumed = nwrite;
		return (SEND_PKT_SENT);
	} else {
		*consumed = 0;
		warn("plugin_send: send returned %d", nwrite);
		plugin_report(pl, clid, REPORT_ERROR_SEND);
		return (SEND_ERROR);
	}
}

void
plugin_custom_receive(struct plugin *pl, uint8_t clid,
		      char *data, int len)
{
	/* nothing to do */
}
