/*-
 * Copyright (c) 2007 Matus Harvan
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *     * The name of the author may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/sysctl.h>

#include <arpa/inet.h>
#include <err.h>
#include <sysexits.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>
#include <event.h>

#include <netinet/in.h>
#include <netinet/udp.h>

#include "mtund.h"
#include "plugin.h"

#define BUFLEN PACKETLEN
#define MAXTMPCONNS 10
#define TEMP_CONN_TIMEOUT 5

enum {
	CONN_STATUS_FREE,	/* connection not in use */
	CONN_STATUS_TEMP,	/* temporary connection */
	CONN_STATUS_PERM	/* permanenet connection */
};

#define min(a,b) ( (a>b) ? b : a )

struct conn {
	int status;
	uint8_t clid;
	int fd;
	struct event ev;
	struct event timer_ev;
	struct plugin_udpall_data *data;
};

struct plugin_udpall_data {
	int state; /* is a client connected? */
	//struct conn conn; /* server - bind fd, client - connect fd */
	struct conn conns[MAXCLIENTS]; /* server - accept fds */
	struct conn tmpconns[MAXTMPCONNS]; /* server - accept fds */
	struct conn *conn;
	struct plugin *pl;
	/* previous value of net.inet.raw.udp_catchall sysctl variable */
	int		 old_sysctl;
	size_t		 old_sysctl_size;
	/* last connection - addresses */
	struct sockaddr_in sa1;
	struct sockaddr_in sa2;
};

void plugin_accept_new_conn(int fd, short ev_type, void *arg);

static void
conn_discard(struct conn *conn)
{
	if (conn->fd != -1) {
		close(conn->fd);
		event_del(&conn->ev);
		if (conn->status == CONN_STATUS_TEMP)
			evtimer_del(&conn->timer_ev);
	}
	conn->fd = -1;
	conn->status = CONN_STATUS_FREE;
}

void
plugin_conn_timeout(int fd, short ev_type, void *arg)
{
	conn_discard(arg);
}

int
plugin_register(struct plugin* pl)
{
    pl->name = "udp_catchall";
    pl->mtu = 1450;
    pl->initialize = plugin_initialize;
    pl->deinitialize = plugin_deinitialize;
    pl->is_ready_to_send = plugin_is_ready_to_send;
    pl->send = plugin_send;
    pl->receive = plugin_receive;
    pl->conn_close = plugin_conn_close;
    pl->conn_map = plugin_conn_map;
    pl->custom_receive = plugin_custom_receive;

    pl->data = malloc(sizeof(struct plugin_udpall_data));
    if (!pl->data)
	    return -1;
    struct plugin_udpall_data *data = pl->data;
    data->pl = pl;
    data->state = PLUGIN_STATE_UNINITIALIZED;
    return 0;
}

void
plugin_unregister(struct plugin* pl)
{
    if (pl->data) {
	free(pl->data);
	pl->data = NULL;
    }
}

/*
 * server: 0 - client, 1 - server
 * 
 * Note that udp_connect wil not fail even if the remote endpoint does
 * not work. It will fail once we try to send data, so it may be wise
 * to first send some handshake data.
 */
int
plugin_initialize(struct plugin *pl, char *host, char *port)
{
    struct plugin_udpall_data *data = pl->data;
    struct sockaddr_in sa;
    int new_sysctl = 1;
    int i;

    if (!server) { /* client */
	    warnx("UDP catchall plugin is server-only!");
	    return -1;
    }

		data->old_sysctl_size = sizeof(data->old_sysctl);
		if (0 != sysctlbyname("net.inet.raw.udp_catchall",
			     &data->old_sysctl, &data->old_sysctl_size,
			     &new_sysctl, sizeof(new_sysctl)))
			errx(EX_UNAVAILABLE, "Cannot set net.inet.raw."
			    "udp_catchall sysctl. Maybe you need to patch "
			    "your kernel?");

    for (i = 0; i < MAXCLIENTS; i++) {
	    data->conns[i].status = CONN_STATUS_FREE;
	    data->conns[i].fd = -1;
	    data->conns[i].data = data;
    }
    for (i = 0; i < MAXTMPCONNS; i++) {
	    data->tmpconns[i].status = CONN_STATUS_FREE;
	    data->tmpconns[i].fd = -1;
	    data->tmpconns[i].data = data;
    }
    data->conn = NULL;
    data->pl = pl;    

    /* create a (catchall) raw UDP socket */
    memset(&sa, 0, sizeof(sa));
    sa.sin_addr.s_addr = INADDR_ANY;
    data->conns->fd = socket(AF_INET, SOCK_RAW, IPPROTO_UDP);
    if (data->conns->fd == -1)
        err(EX_OSERR, "Failed to create a raw UDP socket");

    event_set(&data->conns->ev, data->conns->fd, EV_PERSIST | EV_READ,
	      plugin_accept_new_conn, data->conns);
    if (data->conns->fd != -1) {
	    data->state = PLUGIN_STATE_INITIALIZED;
	    event_add(&data->conns->ev, NULL);
    }
    data->conns->status = CONN_STATUS_PERM;

    if (data->conns->fd != -1) {
	    return 0;
    } else
	    return -1;
}

void
plugin_deinitialize(struct plugin *pl)
{
	int i;
	struct plugin_udpall_data *data = pl->data;
	
	/* deinitialize all connections */
	for (i = 0; i < MAXCLIENTS; i++)
		conn_discard(&data->conns[i]);
	for (i = 0; i < MAXTMPCONNS; i++)
		conn_discard(&data->tmpconns[i]);
	data->conn = NULL;

	sysctlbyname("net.inet.raw.udp_catchall", NULL, 0,
		     &data->old_sysctl, data->old_sysctl_size);

	data->state = PLUGIN_STATE_UNINITIALIZED;
}


void
plugin_conn_close(struct plugin *pl, uint8_t clid)
{
	struct plugin_udpall_data *data = pl->data;

	if (clid == 0)
		conn_discard(data->conn);
	else
		conn_discard(&data->conns[clid]);
}

static void
conn_update_to_perm(struct plugin_udpall_data *data, uint8_t clid, int fd)
{
	struct conn *conn = &data->conns[clid];
	
	printf("conn_update_to_perm()\n");
	if (conn->clid == clid && conn->fd == fd){
		printf("nothing to do...\n");
		return; /* nothing to do */
	}
	
	/* discard previous connection */
	if (conn->status == CONN_STATUS_PERM)
		conn_discard(conn);
	
	/* set up the new connection */
	conn->fd = fd;
	conn->clid = clid;
	conn->status = CONN_STATUS_PERM;
	event_set(&conn->ev, conn->fd, EV_PERSIST | EV_READ,
	    plugin_receive, conn);
	event_add(&conn->ev, NULL);
	evtimer_set(&conn->timer_ev, plugin_conn_timeout, conn);

	data->conn = conn;
}


void
plugin_conn_map(struct plugin *pl, uint8_t clid, int conn_flag)
{
	struct plugin_udpall_data *data = pl->data;
	struct conn *conn = data->conn;
	int fd;
	int new_fd;
	struct timeval tv;
	int on;

	printf("plugin_conn_map()\n");
	
	if (conn->status == CONN_STATUS_FREE) {	/* new connection */
		printf("setting up a new connection\n");
		switch (conn_flag) {
		case CONN_DISCARD:
			return;
			
		case CONN_TEMP:
			break;
			
		case CONN_PERM:
			conn = &data->conns[clid];
			break;

		default:
			warnx("plugin_udp_catchall: invalid conn_flag 0x%x\n",
			    conn_flag);
			return;
		}
		
		/* set up the socket */
		new_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (new_fd == -1) {
			warn("socket() failed");
			return;
		}
		if (0 != setsockopt(new_fd, SOL_SOCKET, SO_REUSEPORT,
			     &on, sizeof(on)))
			warn("setsockopt(SO_REUSEPORT) failed");
		/* set the source port */
                if (0 != bind(new_fd, (struct sockaddr *)&data->sa1,
			      sizeof(struct sockaddr))) {
			warn("bind() failed");
			close(new_fd);
			return;
		}
		/* set the destination address and port */
		if (0 != connect(new_fd, (struct sockaddr *)&data->sa2,
				 sizeof(struct sockaddr))) {
			warn("connect() failed");
			close(new_fd);
			return;
		}
		printf("new client connection accepted\n");

		conn->fd = new_fd;
		event_set(&conn->ev, conn->fd, EV_PERSIST | EV_READ,
			  plugin_receive, conn);
		event_add(&conn->ev, NULL);
		evtimer_set(&conn->timer_ev, plugin_conn_timeout, conn);
		if (conn_flag == CONN_TEMP) {
			conn->clid = 0;
			conn->status = CONN_STATUS_TEMP;
			
			tv.tv_sec = TEMP_CONN_TIMEOUT;
			tv.tv_usec = 0;
			evtimer_add(&conn->timer_ev, &tv);
		} else { /* CONN_PERM */
			conn->clid = clid;
			conn->status = CONN_STATUS_PERM;
			data->conn = conn;
		}

	} else { /* existing connection - update the connection status */
		printf("updating an existing connection\n");
		switch (conn_flag) {
		case CONN_DISCARD:
			conn_discard(conn);
			return;
			
		case CONN_TEMP:
			break;
			
		case CONN_PERM:
			/* migrate a temporary connection to a permanent one */
			fd = conn->fd;
			if (conn->status == CONN_STATUS_TEMP) {
				conn->status = CONN_STATUS_FREE;
				evtimer_del(&conn->timer_ev);
				event_del(&conn->ev);
				conn->fd = -1;
			}
			conn_update_to_perm(data, clid, fd);
			break;
			
		default:
			warnx("plugin_udp_catchall: invalid conn_flag 0x%x\n",
			    conn_flag);
			return;
		}
	}
}

void
plugin_accept_new_conn(int fd, short ev_type, void *arg)
{
	struct plugin_udpall_data *data = ((struct conn *)arg)->data;
	struct conn *conn = NULL;
	int i;
	struct ip *iphdr;
	struct udphdr *uhdr;    
	struct sockaddr_storage from;
	socklen_t fromlen = sizeof(from);
	char buf[BUFLEN];
	int buflen = 0;
	char *bufp = buf;
	char *payload;    
	int payloadlen;
	int nread;
	uint8_t clid;
	int conn_flag;

	printf("plugin_accept_new_conn()\n");
	/* find a free tmpconn to store the connection metadata */
	for (i = 0; i < MAXTMPCONNS; i++)
		if (data->tmpconns[i].status == CONN_STATUS_FREE) {
			conn = &data->tmpconns[i];
			break;
		}
	if (conn == NULL) {
		printf("plugin_udp_catchall: too many new connections\n");
		return;
	}
	data->conn = conn;

	/* read the packet */
        nread = recvfrom(data->conns->fd, buf, BUFLEN, 0,
                         (struct sockaddr *) &from, &fromlen);
	printf("recvfrom() returned %d\n", nread);
	if (nread == -1) {
		plugin_report(data->pl, 0, REPORT_ERROR_RECEIVE);
		return;
	}
	if (nread < sizeof(struct udphdr))
		/* payload too short to contain a UDP header */
		return; 
	buflen = nread;
        /* parse the headers */
	/* IP header */
	iphdr = (struct ip *)bufp;
	bufp += 20;
	buflen -= 20;

	/* UDP header */
	uhdr = (struct udphdr *)bufp;
	bufp += sizeof(*uhdr);
	buflen -= sizeof(*uhdr);

	memset(&data->sa1,0,sizeof(data->sa1));
	data->sa1.sin_port = uhdr->uh_dport;
	memcpy(&data->sa2, &from, sizeof(data->sa2));
	data->sa2.sin_port = uhdr->uh_sport;

	if (buflen <= 0)
		return;

	/* payload */
	payload = bufp;
	payloadlen = buflen;

	process_data_from_plugin(data->pl, payload, payloadlen,
	    &clid, &conn_flag);
	/* conn_map() will se up the connection */
}


void plugin_receive(int fd, short ev_type, void *arg) {
	struct conn *conn = arg;
	struct plugin_udpall_data *data = conn->data;
	int n = 0;
	char packet[PACKETLEN];
	int conn_flag;
	uint8_t clid = 0;

	printf("plugin_receive()\n");

	n = recv(fd, packet, sizeof(packet), 0);
	
	if (n == -1) {
		/* connection broke down or client disconnected */
		conn_discard(conn);
		return;
	}
	
	/* update the most recent connection */
	data->conn = arg;
	
	/* pass the packet to the daemon */
	process_data_from_plugin(data->pl, packet, n, &clid, &conn_flag);
}

int
plugin_is_ready_to_send(struct plugin *pl, uint8_t clid)
{
	return WILL_SEND_IMMEDIATELY;
}

int
plugin_send(struct plugin *pl, uint8_t clid,
    char *data, int len, int data_type, int* consumed)
{
	struct plugin_udpall_data *pldata = pl->data;
	struct conn *conn ;
	int nwrite = 0;
	
	/* locate the right connection */
	if (clid == 0)
		conn = pldata->conn;
	else
		conn = &pldata->conns[clid];
	
	/* send the data */
	nwrite = send(conn->fd, data, len, 0);

	/* error checking */
	if (nwrite == len) {
		*consumed = nwrite;
		return (SEND_PKT_SENT);
	} else {
		*consumed = 0;
		warn("plugin_send: send returned %d", nwrite);
		plugin_report(pl, clid, REPORT_ERROR_SEND);
		return (SEND_ERROR);
	}
}

void
plugin_custom_receive(struct plugin *pl, uint8_t clid,
		      char *data, int len)
{
}
