#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#define SOCKET_TIMEOUT 10

/*
 * Create a listening TCP endpoint. First get the list of potential
 * network layter addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create a
 * listening TCP endpoint is successful (or no other alternative
 * exists).
 */

static int
tcp_listen(char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, on = 1;
    
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    //hints.ai_family = AF_UNSPEC;
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(NULL, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "getaddrinfo failed: %s\n",
                gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }

        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
        if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        fprintf(stderr, "bind failed for port %s\n",
                port);
        return -1;
    }

    if (listen(fd, 1) < 0) {
        fprintf(stderr, "listen failed: %s\n", strerror(errno));
	close(fd);
        return -1;
    }

    return fd;
}

/*
 * Accept a new TCP connection and write a message about who was
 * accepted to the system log.
 */

static int
tcp_accept(int listen)
{
    struct sockaddr_storage ss;
    socklen_t ss_len = sizeof(ss);
    char host[NI_MAXHOST];
    char serv[NI_MAXSERV];
    int n, fd;
    fd = accept(listen, (struct sockaddr *) &ss, &ss_len);
    if (fd == -1) {
        syslog(LOG_ERR, "accept failed: %s", strerror(errno));
        return -1;
    }

    n = getnameinfo((struct sockaddr *) &ss, ss_len,
                    host, sizeof(host), serv, sizeof(serv),
                    NI_NUMERICHOST);
    if (n) {
        fprintf(stderr, "getnameinfo failed: %s", gai_strerror(n));
    } else {
        fprintf(stderr, "connection from %s:%s", host, serv);
    }

    return fd;
}

/*
 * Establish a connection to a remote TCP server. First get the list
 * of potential network layer addresses and transport layer port
 * numbers. Iterate through the returned address list until an attempt
 * to establish a TCP connection is successful (or no other
 * alternative exists).
 */

static int
tcp_connect(const char *host, const char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, serrs = 0, cerrs = 0;
    struct timeval tv;
    
    memset(&hints, 0, sizeof(hints));
    //hints.ai_family = AF_UNSPEC;
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(host, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            serrs++;
            continue;
        }

	tv.tv_sec  = SOCKET_TIMEOUT;
	tv.tv_usec = 0;
	setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv));

        if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        cerrs++;
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        if ((serrs == 1 && ! cerrs) || (!serrs && cerrs == 1)) {
            fprintf(stderr, "%s: %s\n",
                    serrs ? "socket" : "connect", strerror(errno));
        } else {
            fprintf(stderr, "socket or connect: failed for %s port %s\n",
                    host, port);
        }
        return -1;
    }

    return fd;
}
/*
 * Close a TCP connection. This function trivially calls close() on
 * POSIX systems, but might be more complicated on other systems.
 */

static int
tcp_close(int fd)
{
    return close(fd);
}


int
main()
{
    int fd;
    int new_fd;
    char *msg = "Welcome to listenalld\r\n";
    char buf[1234];
    int n, nwrite, nread;
    int soval = 1;
    int count = 50;
    
    fd = tcp_listen("1234");

    if (0 != setsockopt(fd, IPPROTO_TCP, TCP_LISTENALL, &soval, sizeof(soval)))
	err(EX_UNAVAILABLE, "setsockopt(TCP_LISTENALL) failed");
    
    while (count > 0) {
	new_fd = tcp_accept(fd);
	if (new_fd == -1) {
	    fprintf(stderr, "accept failed\n");
	} else {
	    fprintf(stderr, "new client connection accepted\n");
	    n = 0;
		nwrite = 0;
		do {
			nwrite = send(new_fd, msg + n, strlen(msg) - n, 0);
			n += nwrite;
		} while (n < strlen(msg) && nwrite > 0);
		nread = 0;
		do {
			nread = recv(new_fd, buf, sizeof(buf), 0);
			if (nread != -1) {
				buf[nread] = 0;
				nwrite = send(new_fd, buf, nread, 0);
			}
		} while (nread != -1 &&
		      strncmp("exit", buf, strlen("exit")) != 0);
		tcp_close(new_fd);
	}
	count--;
    }
    tcp_close(fd);
    return 0;
}
